"""Defines wandb.init() and associated classes and methods.

`wandb.init()` indicates the beginning of a new run. In an ML training pipeline,
you could add `wandb.init()` to the beginning of your training script as well as
your evaluation script, and each step would be tracked as a run in W&B.

For more on using `wandb.init()`, including code snippets, check out our
[guide and FAQs](https://docs.wandb.ai/guides/track/launch).
"""

from __future__ import annotations

import contextlib
import dataclasses
import functools
import json
import logging
import os
import pathlib
import platform
import sys
import tempfile
import time
from typing import TYPE_CHECKING, Iterable, Iterator, Sequence

from typing_extensions import Any, Literal, Protocol, Self

import wandb
import wandb.env
from wandb import env, trigger
from wandb.analytics import get_sentry
from wandb.errors import CommError, Error, UsageError
from wandb.errors.links import url_registry
from wandb.errors.util import ProtobufErrorHandler
from wandb.integration import sagemaker, weave
from wandb.proto.wandb_telemetry_pb2 import Deprecated
from wandb.sdk.lib import ipython as wb_ipython
from wandb.sdk.lib import progress, runid, wb_logging
from wandb.sdk.lib.paths import StrPath
from wandb.util import _is_artifact_representation

from . import wandb_login, wandb_setup
from .backend.backend import Backend
from .lib import SummaryDisabled, filesystem, module, paths, printer, telemetry
from .lib.deprecation import UNSET, DoNotSet, warn_and_record_deprecation
from .mailbox import wait_with_progress
from .wandb_helper import parse_config
from .wandb_run import Run, TeardownHook, TeardownStage
from .wandb_settings import Settings

if TYPE_CHECKING:
    import wandb.jupyter


def _huggingface_version() -> str | None:
    if "transformers" in sys.modules:
        trans = wandb.util.get_module("transformers")
        if hasattr(trans, "__version__"):
            return str(trans.__version__)
    return None


def _handle_launch_config(settings: Settings) -> dict[str, Any]:
    launch_run_config: dict[str, Any] = {}
    if not settings.launch:
        return launch_run_config
    if os.environ.get("WANDB_CONFIG") is not None:
        try:
            launch_run_config = json.loads(os.environ.get("WANDB_CONFIG", "{}"))
        except (ValueError, SyntaxError):
            wandb.termwarn("Malformed WANDB_CONFIG, using original config")
    elif settings.launch_config_path and os.path.exists(settings.launch_config_path):
        with open(settings.launch_config_path) as fp:
            launch_config = json.loads(fp.read())
        launch_run_config = launch_config.get("overrides", {}).get("run_config")
    else:
        i = 0
        chunks = []
        while True:
            key = f"WANDB_CONFIG_{i}"
            if key in os.environ:
                chunks.append(os.environ[key])
                i += 1
            else:
                break
        if len(chunks) > 0:
            config_string = "".join(chunks)
            try:
                launch_run_config = json.loads(config_string)
            except (ValueError, SyntaxError):
                wandb.termwarn("Malformed WANDB_CONFIG, using original config")

    return launch_run_config


@dataclasses.dataclass(frozen=True)
class _ConfigParts:
    base_no_artifacts: dict[str, Any]
    """The run config passed to `init()` minus any artifact-valued keys."""

    sweep_no_artifacts: dict[str, Any]
    """The config loaded as part of a sweep minus any artifact-valued keys."""

    launch_no_artifacts: dict[str, Any]
    """The config loaded as part of Launch minus any artifact-valued keys."""

    artifacts: dict[str, Any]
    """Artifact keys removed from config dictionaries.

    Due to implementation details of how a Run is constructed,
    artifacts must be inserted into its config after initialization.
    """


class _PrinterCallback(Protocol):
    """A callback for displaying messages after a printer is configured.

    This is used for a few messages that may be generated before run settings
    are computed, which are necessary for creating a printer.
    """

    def __call__(self, run_printer: printer.Printer) -> None:
        """Display information through the given printer."""


def _noop_printer_callback() -> _PrinterCallback:
    """A printer callback that does not print anything."""
    return lambda _: None


def _concat_printer_callbacks(
    cbs: Iterable[_PrinterCallback],
) -> _PrinterCallback:
    """Returns a printer callback that runs the given callbacks in order."""

    def do_callbacks(run_printer: printer.Printer) -> None:
        for cb in cbs:
            cb(run_printer)

    return do_callbacks


class _WandbInit:
    def __init__(
        self,
        wl: wandb_setup._WandbSetup,
        telemetry: telemetry.TelemetryRecord,
    ) -> None:
        self._wl = wl

        self._telemetry = telemetry
        """Telemetry gathered before creating a run.

        After the run is created, `telemetry.context()` is used instead.
        """

        self.kwargs = None
        self.run: Run | None = None
        self.backend: Backend | None = None

        self._teardown_hooks: list[TeardownHook] = []
        self.notebook: wandb.jupyter.Notebook | None = None

        self.deprecated_features_used: list[tuple[Deprecated, str]] = []

    @property
    def _logger(self) -> wandb_setup.Logger:
        return self._wl._get_logger()

    def maybe_login(self, init_settings: Settings) -> None:
        """Log in if we are not creating an offline or disabled run.

        This may change the W&B singleton settings.

        Args:
            init_settings: Settings passed to `wandb.init()` or set via
                keyword arguments.
        """
        # Allow settings passed to init() to override inferred values.
        #
        # Calling login() may change settings on the singleton,
        # so these may not be the final run settings.
        run_settings = self._wl.settings.model_copy()
        run_settings.update_from_settings(init_settings)

        # NOTE: _noop or _offline can become true after _login().
        #   _noop happens if _login hits a timeout.
        #   _offline can be selected by the user at the login prompt.
        if run_settings._noop or run_settings._offline:
            return

        wandb_login._login(
            host=run_settings.base_url,
            force=run_settings.force,
            key=run_settings.api_key,
            # Do not save an explicitly provided API key to .netrc.
            update_api_key=run_settings.api_key is None,
            _silent=run_settings.quiet or run_settings.silent,
        )

    def warn_env_vars_change_after_setup(self) -> _PrinterCallback:
        """Warn if environment variables changed after `wandb.setup()`.

        Returns:
            A callback to print any generated warnings.
        """
        if not self._wl.did_environment_change():
            return _noop_printer_callback()

        def print_warning(run_printer: printer.Printer) -> None:
            line = (
                "Changes to your `wandb` environment variables will be ignored "
                "because your `wandb` session has already started. "
                "For more information on how to modify your settings with "
                "`wandb.init()` arguments, please refer to "
                f"{run_printer.link(url_registry.url('wandb-init'), 'the W&B docs')}."
            )
            run_printer.display(line, level="warn")

        return print_warning

    def clear_run_path_if_sweep_or_launch(
        self,
        init_settings: Settings,
    ) -> _PrinterCallback:
        """Clear project/entity/run_id keys if in a Sweep or a Launch context.

        Args:
            init_settings: Settings specified in the call to `wandb.init()`.

        Returns:
            A callback to print any generated warnings.
        """
        when_doing_thing = ""

        if self._wl.settings.sweep_id:
            when_doing_thing = "when running a sweep"
        elif self._wl.settings.launch:
            when_doing_thing = "when running from a wandb launch context"

        if not when_doing_thing:
            return _noop_printer_callback()

        warnings = []

        def warn(key: str, value: str) -> None:
            warnings.append(f"Ignoring {key} {value!r} {when_doing_thing}.")

        if init_settings.project is not None:
            warn("project", init_settings.project)
            init_settings.project = None
        if init_settings.entity is not None:
            warn("entity", init_settings.entity)
            init_settings.entity = None
        if init_settings.run_id is not None:
            warn("run_id", init_settings.run_id)
            init_settings.run_id = None

        def print_warnings(run_printer: printer.Printer) -> None:
            for warning in warnings:
                run_printer.display(warning, level="warn")

        return print_warnings

    def make_run_settings(
        self,
        init_settings: Settings,
    ) -> tuple[Settings, _PrinterCallback]:
        """Returns the run's settings and any warnings.

        Args:
            init_settings: Settings passed to `wandb.init()` or set via
                keyword arguments.
        """
        warning_callbacks: list[_PrinterCallback] = [
            self.warn_env_vars_change_after_setup(),
            self.clear_run_path_if_sweep_or_launch(init_settings),
        ]

        # Inherit global settings.
        settings = self._wl.settings.model_copy()

        # Apply settings from wandb.init() call.
        settings.update_from_settings(init_settings)

        # Infer the run ID from SageMaker.
        if not settings.sagemaker_disable and sagemaker.is_using_sagemaker():
            if sagemaker.set_run_id(settings):
                self._logger.info("set run ID and group based on SageMaker")
                self._telemetry.feature.sagemaker = True

        # get status of code saving before applying user settings
        save_code_pre_user_settings = settings.save_code
        if not settings._offline and not settings._noop:
            user_settings = self._wl._load_user_settings()
            if user_settings is not None:
                settings.update_from_dict(user_settings)

        # ensure that user settings don't set saving to true
        # if user explicitly set these to false in UI
        if save_code_pre_user_settings is False:
            settings.save_code = False

        # TODO: remove this once we refactor the client. This is a temporary
        # fix to make sure that we use the same project name for wandb-core.
        # The reason this is not going through the settings object is to
        # avoid failure cases in other parts of the code that will be
        # removed with the switch to wandb-core.
        if settings.project is None:
            settings.project = wandb.util.auto_project_name(settings.program)

        settings.x_start_time = time.time()

        # In shared mode, generate a unique label if not provided.
        # The label is used to distinguish between system metrics and console logs
        # from different writers to the same run.
        if settings._shared and not settings.x_label:
            # TODO: If executed in a known distributed environment (e.g. Ray or SLURM),
            #   use the env vars to generate a label (e.g. SLURM_JOB_ID or RANK)
            prefix = settings.host or ""
            label = runid.generate_id()
            settings.x_label = f"{prefix}-{label}" if prefix else label

        return settings, _concat_printer_callbacks(warning_callbacks)

    def _load_autoresume_run_id(self, resume_file: pathlib.Path) -> str | None:
        """Returns the run_id stored in the auto-resume file, if any.

        Returns `None` if the file does not exist or is not in a valid format.

        Args:
            resume_file: The file path to use for resume='auto' mode.
        """
        if not resume_file.exists():
            return None

        with resume_file.open() as f:
            try:
                return json.load(f)["run_id"]

            except json.JSONDecodeError as e:
                self._logger.exception(
                    f"could not decode {resume_file}, ignoring",
                    exc_info=e,
                )
                return None

            except KeyError:
                self._logger.exception(
                    f"resume file at {resume_file} did not store a run_id"
                )
                return None

    def _save_autoresume_run_id(
        self,
        *,
        resume_file: pathlib.Path,
        run_id: str,
    ) -> None:
        """Write the run ID to the auto-resume file."""
        resume_file.parent.mkdir(exist_ok=True)
        with resume_file.open("w") as f:
            json.dump({"run_id": run_id}, f)

    def set_run_id(self, settings: Settings) -> None:
        """Set the run ID and possibly save it to the auto-resume file.

        After this, `settings.run_id` is guaranteed to be set.

        If a `resume_from` is provided and `run_id` is not set, initialize
        `run_id` with the `resume_from` run's `run_id`.

        Args:
            settings: The run's settings derived from the environment
                and explicit values passed to `wandb.init()`.
        """
        if settings.resume == "auto" and settings.resume_fname:
            resume_path = pathlib.Path(settings.resume_fname)
        else:
            resume_path = None

        if resume_path:
            previous_id = self._load_autoresume_run_id(resume_path)

            if not previous_id:
                pass
            elif settings.run_id is None:
                self._logger.info(f"loaded run ID from {resume_path}")
                settings.run_id = previous_id
            elif settings.run_id != previous_id:
                wandb.termwarn(
                    f"Ignoring ID {previous_id} loaded due to resume='auto'"
                    f" because the run ID is set to {settings.run_id}.",
                )

        # If no run ID was inferred, explicitly set, or loaded from an
        # auto-resume file, then we generate a new ID.
        if settings.run_id is None:
            # If resume_from is provided and run_id is not already set,
            # initialize run_id with the value from resume_from.
            if settings.resume_from:
                settings.run_id = settings.resume_from.run
            else:
                settings.run_id = runid.generate_id()

        if resume_path:
            self._save_autoresume_run_id(
                resume_file=resume_path,
                run_id=settings.run_id,
            )

    def set_sync_dir_suffix(self, settings: Settings) -> None:
        """Add a suffix to sync_dir if it already exists.

        The sync_dir uses a timestamp with second-level precision which can
        result in conflicts if a run with the same ID is initialized within the
        same second. This is most likely to happen in tests.

        This can't prevent conflicts from multiple processes attempting
        to create a wandb run simultaneously.

        Args:
            settings: Fully initialized settings other than the
                x_sync_dir_suffix setting which will be modified.
        """
        index = 1
        while pathlib.Path(settings.sync_dir).exists():
            settings.x_sync_dir_suffix = f"{index}"
            index += 1

    def make_run_config(
        self,
        settings: Settings,
        config: dict | str | None = None,
        config_exclude_keys: list[str] | None = None,
        config_include_keys: list[str] | None = None,
    ) -> _ConfigParts:
        """Construct the run's config.

        Args:
            settings: The run's finalized settings.
            config: The config passed to `init()`.
            config_exclude_keys: Deprecated. Keys to filter out from `config`.
            config_include_keys: Deprecated. Keys to include from `config`.

        Returns:
            Initial values for the run's config.
        """
        if config_exclude_keys:
            self.deprecated_features_used.append(
                (
                    Deprecated(init__config_exclude_keys=True),
                    "config_exclude_keys is deprecated. Use"
                    " `config=wandb.helper.parse_config(config_object,"
                    " exclude=('key',))` instead.",
                )
            )
        if config_include_keys:
            self.deprecated_features_used.append(
                (
                    Deprecated(init__config_include_keys=True),
                    "config_include_keys is deprecated. Use"
                    " `config=wandb.helper.parse_config(config_object,"
                    " include=('key',))` instead.",
                )
            )
        config = parse_config(
            config or dict(),
            include=config_include_keys,
            exclude=config_exclude_keys,
        )

        result = _ConfigParts(
            base_no_artifacts=dict(),
            sweep_no_artifacts=dict(),
            launch_no_artifacts=dict(),
            artifacts=dict(),
        )

        if not settings.sagemaker_disable and sagemaker.is_using_sagemaker():
            sagemaker_config = sagemaker.parse_sm_config()
            self._split_artifacts_from_config(
                sagemaker_config,
                config_target=result.base_no_artifacts,
                artifacts=result.artifacts,
            )
            self._telemetry.feature.sagemaker = True

        if self._wl.config:
            self._split_artifacts_from_config(
                self._wl.config,
                config_target=result.base_no_artifacts,
                artifacts=result.artifacts,
            )

        if config and isinstance(config, dict):
            self._split_artifacts_from_config(
                config,
                config_target=result.base_no_artifacts,
                artifacts=result.artifacts,
            )

        if self._wl._sweep_config:
            self._split_artifacts_from_config(
                self._wl._sweep_config,
                config_target=result.sweep_no_artifacts,
                artifacts=result.artifacts,
            )

        if launch_config := _handle_launch_config(settings):
            self._split_artifacts_from_config(
                launch_config,
                config_target=result.launch_no_artifacts,
                artifacts=result.artifacts,
            )

        wandb_internal = result.base_no_artifacts.setdefault("_wandb", dict())

        if settings.save_code and settings.program_relpath:
            wandb_internal["code_path"] = paths.LogicalPath(
                os.path.join("code", settings.program_relpath)
            )
        if settings.fork_from is not None:
            wandb_internal["branch_point"] = {
                "run_id": settings.fork_from.run,
                "step": settings.fork_from.value,
            }
        if settings.resume_from is not None:
            wandb_internal["branch_point"] = {
                "run_id": settings.resume_from.run,
                "step": settings.resume_from.value,
            }

        return result

    def teardown(self) -> None:
        # TODO: currently this is only called on failed wandb.init attempts
        # normally this happens on the run object
        self._logger.info("tearing down wandb.init")
        for hook in self._teardown_hooks:
            hook.call()

    def _split_artifacts_from_config(
        self,
        config_source: dict,
        config_target: dict,
        artifacts: dict,
    ) -> None:
        for k, v in config_source.items():
            if _is_artifact_representation(v):
                artifacts[k] = v
            else:
                config_target.setdefault(k, v)

    def _safe_symlink(
        self, base: str, target: str, name: str, delete: bool = False
    ) -> None:
        # TODO(jhr): do this with relpaths, but i can't figure it out on no sleep
        if not hasattr(os, "symlink"):
            return

        pid = os.getpid()
        tmp_name = os.path.join(base, f"{name}.{pid}")

        if delete:
            try:
                os.remove(os.path.join(base, name))
            except OSError:
                pass
        target = os.path.relpath(target, base)
        try:
            os.symlink(target, tmp_name)
            os.rename(tmp_name, os.path.join(base, name))
        except OSError:
            pass

    def _pre_run_cell_hook(self, *args, **kwargs) -> None:
        """Hook for the IPython pre_run_cell event.

        This pauses a run, preventing system metrics from being collected
        the run's runtime from increasing. It also uploads the notebook's code.
        """
        if not self.backend:
            return

        if self.notebook and self.notebook.save_ipynb():
            assert self.run is not None
            res = self.run.log_code(root=None)
            self._logger.info("saved code: %s", res)

        if self.backend.interface is not None:
            self._logger.info("pausing backend")
            self.backend.interface.publish_pause()

    def _post_run_cell_hook(self, *args, **kwargs) -> None:
        """Hook for the IPython post_run_cell event.

        Resumes collection of system metrics and the run's timer.
        """
        if self.backend is None or self.backend.interface is None:
            return

        self._logger.info("resuming backend")
        self.backend.interface.publish_resume()

    def _jupyter_teardown(self) -> None:
        """Teardown hooks and display saving, called with wandb.finish."""
        assert self.notebook
        ipython = self.notebook.shell

        if self.run:
            self.notebook.save_history(self.run)

        if self.notebook.save_ipynb():
            assert self.run is not None
            res = self.run.log_code(root=None)
            self._logger.info("saved code and history: %s", res)
        self._logger.info("cleaning up jupyter logic")

        ipython.events.unregister("pre_run_cell", self._pre_run_cell_hook)
        ipython.events.unregister("post_run_cell", self._post_run_cell_hook)

        ipython.display_pub.publish = ipython.display_pub._orig_publish
        del ipython.display_pub._orig_publish

    def monkeypatch_ipython(self, settings: Settings) -> None:
        """Add hooks, and session history saving."""
        self.notebook = wandb.jupyter.Notebook(settings)
        ipython = self.notebook.shell

        # Monkey patch ipython publish to capture displayed outputs
        if not hasattr(ipython.display_pub, "_orig_publish"):
            self._logger.info("configuring jupyter hooks %s", self)
            ipython.display_pub._orig_publish = ipython.display_pub.publish

            ipython.events.register("pre_run_cell", self._pre_run_cell_hook)
            ipython.events.register("post_run_cell", self._post_run_cell_hook)

            self._teardown_hooks.append(
                TeardownHook(self._jupyter_teardown, TeardownStage.EARLY)
            )

        def publish(data, metadata=None, **kwargs) -> None:
            ipython.display_pub._orig_publish(data, metadata=metadata, **kwargs)
            assert self.notebook is not None
            self.notebook.save_display(
                ipython.execution_count, {"data": data, "metadata": metadata}
            )

        ipython.display_pub.publish = publish

    @contextlib.contextmanager
    def setup_run_log_directory(self, settings: Settings) -> Iterator[None]:
        """Set up the run's log directory.

        This is a context manager that closes and unregisters the log handler
        in case of an uncaught exception, so that future logged messages do not
        modify this run's log file.
        """
        filesystem.mkdir_exists_ok(os.path.dirname(settings.log_user))
        filesystem.mkdir_exists_ok(os.path.dirname(settings.log_internal))
        filesystem.mkdir_exists_ok(os.path.dirname(settings.sync_file))
        filesystem.mkdir_exists_ok(settings.files_dir)
        filesystem.mkdir_exists_ok(settings._tmp_code_dir)

        if settings.symlink:
            self._safe_symlink(
                os.path.dirname(settings.sync_symlink_latest),
                os.path.dirname(settings.sync_file),
                os.path.basename(settings.sync_symlink_latest),
                delete=True,
            )
            self._safe_symlink(
                os.path.dirname(settings.log_symlink_user),
                settings.log_user,
                os.path.basename(settings.log_symlink_user),
                delete=True,
            )
            self._safe_symlink(
                os.path.dirname(settings.log_symlink_internal),
                settings.log_internal,
                os.path.basename(settings.log_symlink_internal),
                delete=True,
            )

        assert settings.run_id
        handler = wb_logging.add_file_handler(
            settings.run_id,
            pathlib.Path(settings.log_user),
        )

        if env.is_debug():
            handler.setLevel(logging.DEBUG)

        disposed = False

        def dispose_handler() -> None:
            nonlocal disposed

            if not disposed:
                disposed = True
                logging.getLogger("wandb").removeHandler(handler)
                handler.close()

        try:
            self._teardown_hooks.append(
                TeardownHook(
                    call=dispose_handler,
                    stage=TeardownStage.LATE,
                )
            )

            self._wl._early_logger_flush(logging.getLogger("wandb"))
            self._logger.info(f"Logging user logs to {settings.log_user}")
            self._logger.info(f"Logging internal logs to {settings.log_internal}")

            yield
        except Exception:
            dispose_handler()
            raise

    def make_disabled_run(self, config: _ConfigParts) -> Run:
        """Returns a Run-like object where all methods are no-ops.

        This method is used when the `mode` setting is set to "disabled", such as
        by wandb.init(mode="disabled") or by setting the WANDB_MODE environment
        variable to "disabled".

        It creates a Run object that mimics the behavior of a normal Run but doesn't
        communicate with the W&B servers.

        The returned Run object has all expected attributes and methods, but they
        are no-op versions that don't perform any actual logging or communication.
        """
        run_id = runid.generate_id()
        drun = Run(
            settings=Settings(
                mode="disabled",
                root_dir=tempfile.gettempdir(),
                run_id=run_id,
                run_tags=tuple(),
                run_notes=None,
                run_group=None,
                run_name=f"dummy-{run_id}",
                project="dummy",
                entity="dummy",
            )
        )
        # config, summary, and metadata objects
        drun._config = wandb.sdk.wandb_config.Config()
        drun._config.update(config.sweep_no_artifacts)
        drun._config.update(config.base_no_artifacts)
        drun.summary = SummaryDisabled()  # type: ignore

        # methods
        drun.log = lambda data, *_, **__: drun.summary.update(data)  # type: ignore[method-assign]
        drun.finish = lambda *_, **__: module.unset_globals()  # type: ignore[method-assign]
        drun.join = drun.finish  # type: ignore[method-assign]
        drun.define_metric = lambda *_, **__: wandb.sdk.wandb_metric.Metric("dummy")  # type: ignore[method-assign]
        drun.save = lambda *_, **__: False  # type: ignore[method-assign]
        for symbol in (
            "alert",
            "finish_artifact",
            "get_project_url",
            "get_sweep_url",
            "get_url",
            "link_artifact",
            "link_model",
            "use_artifact",
            "log_code",
            "log_model",
            "use_model",
            "mark_preempting",
            "restore",
            "status",
            "watch",
            "unwatch",
            "upsert_artifact",
            "_finish",
        ):
            setattr(drun, symbol, lambda *_, **__: None)  # type: ignore

        # set properties to None
        for attr in ("url", "project_url", "sweep_url"):
            setattr(type(drun), attr, property(lambda _: None))

        class _ChainableNoOp:
            """An object that allows chaining arbitrary attributes and method calls."""

            def __getattr__(self, _: str) -> Self:
                return self

            def __call__(self, *_: Any, **__: Any) -> Self:
                return self

        class _ChainableNoOpField:
            # This is used to chain arbitrary attributes and method calls.
            # For example, `run.log_artifact().state` will work in disabled mode.
            def __init__(self) -> None:
                self._value = None

            def __set__(self, instance: Any, value: Any) -> None:
                self._value = value

            def __get__(self, instance: Any, owner: type) -> Any:
                return _ChainableNoOp() if (self._value is None) else self._value

            def __call__(self, *args: Any, **kwargs: Any) -> _ChainableNoOp:
                return _ChainableNoOp()

        drun.log_artifact = _ChainableNoOpField()  # type: ignore
        # attributes
        drun._start_time = time.time()
        drun._starting_step = 0
        drun._step = 0
        drun._attach_id = None
        drun._backend = None

        # set the disabled run as the global run
        module.set_global(
            run=drun,
            config=drun.config,
            log=drun.log,
            summary=drun.summary,
            save=drun.save,
            use_artifact=drun.use_artifact,
            log_artifact=drun.log_artifact,
            define_metric=drun.define_metric,
            alert=drun.alert,
            watch=drun.watch,
            unwatch=drun.unwatch,
        )
        return drun

    def init(  # noqa: C901
        self,
        settings: Settings,
        config: _ConfigParts,
        run_printer: printer.Printer,
    ) -> Run:
        self._logger.info("calling init triggers")
        trigger.call("on_init")

        assert self._wl is not None

        self._logger.info(
            f"wandb.init called with sweep_config: {config.sweep_no_artifacts}"
            f"\nconfig: {config.base_no_artifacts}"
        )

        if previous_run := self._wl.most_recent_active_run:
            if (
                settings.reinit in (True, "finish_previous")
                # calling wandb.init() in notebooks finishes previous runs
                # by default for user convenience.
                or (settings.reinit == "default" and wb_ipython.in_notebook())
            ):
                run_printer.display(
                    "Finishing previous runs because reinit is set"
                    f" to {settings.reinit!r}."
                )
                self._wl.finish_all_active_runs()

            elif settings.reinit == "create_new":
                self._logger.info(
                    "wandb.init() called while a run is active,"
                    " and reinit is set to 'create_new', so continuing"
                )

            elif settings.resume == "must":
                raise wandb.Error(
                    "Cannot resume a run while another run is active."
                    " You must either finish it using run.finish(),"
                    " or use reinit='create_new' when calling wandb.init()."
                )

            else:
                run_printer.display(
                    "wandb.init() called while a run is active and reinit is"
                    f" set to {settings.reinit!r}, so returning the previous"
                    " run."
                )

                with telemetry.context(run=previous_run) as tel:
                    tel.feature.init_return_run = True

                return previous_run

        self._logger.info("starting backend")

        service = self._wl.ensure_service()
        self._logger.info("sending inform_init request")
        service.inform_init(
            settings=settings.to_proto(),
            run_id=settings.run_id,  # type: ignore
        )

        backend = Backend(settings=settings, service=service)
        backend.ensure_launched()
        self._logger.info("backend started and connected")

        run = Run(
            config=config.base_no_artifacts,
            settings=settings,
            sweep_config=config.sweep_no_artifacts,
            launch_config=config.launch_no_artifacts,
        )

        # Populate initial telemetry
        with telemetry.context(run=run, obj=self._telemetry) as tel:
            tel.cli_version = wandb.__version__
            tel.python_version = platform.python_version()
            tel.platform = f"{platform.system()}-{platform.machine()}".lower()
            hf_version = _huggingface_version()
            if hf_version:
                tel.huggingface_version = hf_version
            if settings._jupyter:
                tel.env.jupyter = True
            if settings._ipython:
                tel.env.ipython = True
            if settings._colab:
                tel.env.colab = True
            if settings._kaggle:
                tel.env.kaggle = True
            if settings._windows:
                tel.env.windows = True

            if settings.launch:
                tel.feature.launch = True

            for module_name in telemetry.list_telemetry_imports(only_imported=True):
                setattr(tel.imports_init, module_name, True)

            if os.environ.get("PEX"):
                tel.env.pex = True

            if settings._aws_lambda:
                tel.env.aws_lambda = True

            if settings.x_flow_control_disabled:
                tel.feature.flow_control_disabled = True
            if settings.x_flow_control_custom:
                tel.feature.flow_control_custom = True
            if settings._shared:
                wandb.termwarn(
                    "The `shared` mode feature is experimental and may change. "
                    "Please contact support@wandb.com for guidance and to report any issues."
                )
                tel.feature.shared_mode = True

            if settings.x_label:
                tel.feature.user_provided_label = True

            if wandb.env.dcgm_profiling_enabled():
                tel.feature.dcgm_profiling_enabled = True

        if not settings.label_disable:
            if self.notebook:
                run._label_probe_notebook(self.notebook)
            else:
                run._label_probe_main()

        for deprecated_feature, msg in self.deprecated_features_used:
            warn_and_record_deprecation(
                feature=deprecated_feature,
                message=msg,
                run=run,
            )

        self._logger.info("updated telemetry")

        run._set_library(self._wl)
        run._set_backend(backend)
        run._set_teardown_hooks(self._teardown_hooks)

        assert backend.interface
        backend.interface.publish_header()

        # Using GitRepo() blocks & can be slow, depending on user's current git setup.
        # We don't want to block run initialization/start request, so populate run's git
        # info beforehand.
        if not (settings.disable_git or settings.x_disable_machine_info):
            run._populate_git_info()

        if settings._offline and settings.resume:
            wandb.termwarn(
                "`resume` will be ignored since W&B syncing is set to `offline`. "
                f"Starting a new run with run id {run.id}."
            )
        error: wandb.Error | None = None

        timeout = settings.init_timeout

        self._logger.info(
            f"communicating run to backend with {timeout} second timeout",
        )

        run_init_handle = backend.interface.deliver_run(run)

        try:
            with progress.progress_printer(
                run_printer,
                default_text="Waiting for wandb.init()...",
            ) as progress_printer:
                result = wait_with_progress(
                    run_init_handle,
                    timeout=timeout,
                    display_progress=functools.partial(
                        progress.loop_printing_operation_stats,
                        progress_printer,
                        backend.interface,
                    ),
                )

        except TimeoutError:
            run_init_handle.cancel(backend.interface)

            # This may either be an issue with the W&B server (a CommError)
            # or a bug in the SDK (an Error). We cannot distinguish between
            # the two causes here.
            raise CommError(
                f"Run initialization has timed out after {timeout} sec."
                " Please try increasing the timeout with the `init_timeout`"
                " setting: `wandb.init(settings=wandb.Settings(init_timeout=120))`."
            )

        assert result.run_result

        if error := ProtobufErrorHandler.to_exception(result.run_result.error):
            raise error

        if not result.run_result.HasField("run"):
            raise Error("Assertion failed: run_result is missing the run field")

        if result.run_result.run.resumed:
            self._logger.info("run resumed")
            with telemetry.context(run=run) as tel:
                tel.feature.resumed = result.run_result.run.resumed
        run._set_run_obj(result.run_result.run)

        self._logger.info("starting run threads in backend")

        assert backend.interface

        run_start_handle = backend.interface.deliver_run_start(run)
        try:
            # TODO: add progress to let user know we are doing something
            run_start_handle.wait_or(timeout=30)
        except TimeoutError:
            pass

        backend.interface.publish_probe_system_info()

        assert self._wl is not None
        self.run = run

        run._handle_launch_artifact_overrides()
        if (
            settings.launch
            and settings.launch_config_path
            and os.path.exists(settings.launch_config_path)
        ):
            run.save(settings.launch_config_path)
        # put artifacts in run config here
        # since doing so earlier will cause an error
        # as the run is not upserted
        for k, v in config.artifacts.items():
            run.config.update({k: v}, allow_val_change=True)
        job_artifact = run._launch_artifact_mapping.get(
            wandb.util.LAUNCH_JOB_ARTIFACT_SLOT_NAME
        )
        if job_artifact:
            run.use_artifact(job_artifact)

        self.backend = backend

        if settings.reinit != "create_new":
            _set_global_run(run)

        run._on_start()
        self._logger.info("run started, returning control to user process")
        return run


def _attach(
    attach_id: str | None = None,
    run_id: str | None = None,
    *,
    run: Run | None = None,
) -> Run | None:
    """Attach to a run currently executing in another process/thread.

    Args:
        attach_id: (str, optional) The id of the run or an attach identifier
            that maps to a run.
        run_id: (str, optional) The id of the run to attach to.
        run: (Run, optional) The run instance to attach
    """
    attach_id = attach_id or run_id
    if not ((attach_id is None) ^ (run is None)):
        raise UsageError("Either (`attach_id` or `run_id`) or `run` must be specified")

    attach_id = attach_id or (run._attach_id if run else None)

    if attach_id is None:
        raise UsageError(
            "Either `attach_id` or `run_id` must be specified or `run` must have `_attach_id`"
        )

    _wl = wandb_setup.singleton()
    logger = _wl._get_logger()

    service = _wl.ensure_service()

    try:
        attach_settings = service.inform_attach(attach_id=attach_id)
    except Exception as e:
        raise UsageError(f"Unable to attach to run {attach_id}") from e

    settings = _wl.settings.model_copy()
    settings.update_from_dict(
        {
            "run_id": attach_id,
            "x_start_time": attach_settings.x_start_time.value,
            "mode": attach_settings.mode.value,
        }
    )

    # TODO: consolidate this codepath with wandb.init()
    backend = Backend(settings=settings, service=service)
    backend.ensure_launched()
    logger.info("attach backend started and connected")

    if run is None:
        run = Run(settings=settings)
    else:
        run._init(settings=settings)
    run._set_library(_wl)
    run._set_backend(backend)
    assert backend.interface

    attach_handle = backend.interface.deliver_attach(attach_id)
    try:
        # TODO: add progress to let user know we are doing something
        attach_result = attach_handle.wait_or(timeout=30)
    except TimeoutError:
        raise UsageError("Timeout attaching to run")

    attach_response = attach_result.response.attach_response
    if attach_response.error and attach_response.error.message:
        raise UsageError(f"Failed to attach to run: {attach_response.error.message}")

    run._set_run_obj(attach_response.run)
    _set_global_run(run)
    run._on_attach()
    return run


def _set_global_run(run: Run) -> None:
    """Set `wandb.run` and point some top-level functions to its methods.

    Args:
        run: The run to make global.
    """
    module.set_global(
        run=run,
        config=run.config,
        log=run.log,
        summary=run.summary,
        save=run.save,
        use_artifact=run.use_artifact,
        log_artifact=run.log_artifact,
        define_metric=run.define_metric,
        alert=run.alert,
        watch=run.watch,
        unwatch=run.unwatch,
        mark_preempting=run.mark_preempting,
        log_model=run.log_model,
        use_model=run.use_model,
        link_model=run.link_model,
    )


def _monkeypatch_openai_gym() -> None:
    """Patch OpenAI gym to log to the global `wandb.run`."""
    if len(wandb.patched["gym"]) > 0:
        return

    from wandb.integration import gym

    gym.monitor()


def _monkeypatch_tensorboard() -> None:
    """Patch TensorBoard to log to the global `wandb.run`."""
    if len(wandb.patched["tensorboard"]) > 0:
        return

    from wandb.integration import tensorboard as tb_module

    tb_module.patch()


def try_create_root_dir(settings: Settings) -> None:
    """Try to create the root directory specified in settings.

    If creation fails due to permissions or other errors,
    falls back to using the system temp directory.

    Args:
        settings: The runs settings containing root_dir configuration.
            This function may update the root_dir to a temporary directory
            if the parent directory is not writable.
    """
    fallback_to_temp_dir = False

    try:
        os.makedirs(settings.root_dir, exist_ok=True)
    except OSError:
        wandb.termwarn(
            f"Unable to create root directory {settings.root_dir}",
            repeat=False,
        )
        fallback_to_temp_dir = True
    else:
        if not os.access(settings.root_dir, os.W_OK | os.R_OK):
            wandb.termwarn(
                f"Path {settings.root_dir} wasn't read/writable",
                repeat=False,
            )
            fallback_to_temp_dir = True

    if not fallback_to_temp_dir:
        return

    tmp_dir = tempfile.gettempdir()
    if not os.access(tmp_dir, os.W_OK | os.R_OK):
        raise ValueError(
            f"System temp directory ({tmp_dir}) is not writable/readable, "
            "please set the `dir` argument in `wandb.init()` to a writable/readable directory."
        )

    settings.root_dir = tmp_dir
    wandb.termwarn(
        f"Falling back to temporary directory {tmp_dir}.",
        repeat=False,
    )
    os.makedirs(settings.root_dir, exist_ok=True)


def init(  # noqa: C901
    entity: str | None = None,
    project: str | None = None,
    dir: StrPath | None = None,
    id: str | None = None,
    name: str | None = None,
    notes: str | None = None,
    tags: Sequence[str] | None = None,
    config: dict[str, Any] | str | None = None,
    config_exclude_keys: list[str] | None = None,
    config_include_keys: list[str] | None = None,
    allow_val_change: bool | None = None,
    group: str | None = None,
    job_type: str | None = None,
    mode: Literal["online", "offline", "disabled", "shared"] | None = None,
    force: bool | None = None,
    reinit: (
        bool
        | Literal[
            None,
            "default",
            "return_previous",
            "finish_previous",
            "create_new",
        ]
    ) = None,
    resume: bool | Literal["allow", "never", "must", "auto"] | None = None,
    resume_from: str | None = None,
    fork_from: str | None = None,
    save_code: bool | None = None,
    tensorboard: bool | None = None,
    sync_tensorboard: bool | None = None,
    monitor_gym: bool | None = None,
    settings: Settings | dict[str, Any] | None = None,
    anonymous: DoNotSet = UNSET,
) -> Run:
    r"""Start a new run to track and log to W&B.

    In an ML training pipeline, you could add `wandb.init()` to the beginning of
    your training script as well as your evaluation script, and each piece would
    be tracked as a run in W&B.

    `wandb.init()` spawns a new background process to log data to a run, and it
    also syncs data to https://wandb.ai by default, so you can see your results
    in real-time. When you're done logging data, call `wandb.Run.finish()` to end the run.
    If you don't call `run.finish()`, the run will end when your script exits.

    Run IDs must not contain any of the following special characters `/ \ # ? % :`

    Args:
        entity: The username or team name the runs are logged to.
            The entity must already exist, so ensure you create your account
            or team in the UI before starting to log runs. If not specified, the
            run will default your default entity. To change the default entity,
            go to your settings and update the
            "Default location to create new projects" under "Default team".
        project: The name of the project under which this run will be logged.
            If not specified, we use a heuristic to infer the project name based
            on the system, such as checking the git root or the current program
            file. If we can't infer the project name, the project will default to
            `"uncategorized"`.
        dir: The absolute path to the directory where experiment logs and
            metadata files are stored. If not specified, this defaults
            to the `./wandb` directory. Note that this does not affect the
            location where artifacts are stored when calling `download()`.
        id: A unique identifier for this run, used for resuming. It must be unique
            within the project and cannot be reused once a run is deleted. For
            a short descriptive name, use the `name` field,
            or for saving hyperparameters to compare across runs, use `config`.
        name: A short display name for this run, which appears in the UI to help
            you identify it. By default, we generate a random two-word name
            allowing easy cross-reference runs from table to charts. Keeping these
            run names brief enhances readability in chart legends and tables. For
            saving hyperparameters, we recommend using the `config` field.
        notes: A detailed description of the run, similar to a commit message in
            Git. Use this argument to capture any context or details that may
            help you recall the purpose or setup of this run in the future.
        tags: A list of tags to label this run in the UI. Tags are helpful for
            organizing runs or adding temporary identifiers like "baseline" or
            "production." You can easily add, remove tags, or filter by tags in
            the UI.
            If resuming a run, the tags provided here will replace any existing
            tags. To add tags to a resumed run without overwriting the current
            tags, use `run.tags += ("new_tag",)` after calling `run = wandb.init()`.
        config: Sets `wandb.config`, a dictionary-like object for storing input
            parameters to your run, such as model hyperparameters or data
            preprocessing settings.
            The config appears in the UI in an overview page, allowing you to
            group, filter, and sort runs based on these parameters.
            Keys should not contain periods (`.`), and values should be
            smaller than 10 MB.
            If a dictionary, `argparse.Namespace`, or `absl.flags.FLAGS` is
            provided, the key-value pairs will be loaded directly into
            `wandb.config`.
            If a string is provided, it is interpreted as a path to a YAML file,
            from which configuration values will be loaded into `wandb.config`.
        config_exclude_keys: A list of specific keys to exclude from `wandb.config`.
        config_include_keys: A list of specific keys to include in `wandb.config`.
        allow_val_change: Controls whether config values can be modified after their
            initial set. By default, an exception is raised if a config value is
            overwritten. For tracking variables that change during training, such as
            a learning rate, consider using `wandb.log()` instead. By default, this
            is `False` in scripts and `True` in Notebook environments.
        group: Specify a group name to organize individual runs as part of a larger
            experiment. This is useful for cases like cross-validation or running
            multiple jobs that train and evaluate a model on different test sets.
            Grouping allows you to manage related runs collectively in the UI,
            making it easy to toggle and review results as a unified experiment.
        job_type: Specify the type of run, especially helpful when organizing runs
            within a group as part of a larger experiment. For example, in a group,
            you might label runs with job types such as "train" and "eval".
            Defining job types enables you to easily filter and group similar runs
            in the UI, facilitating direct comparisons.
        mode: Specifies how run data is managed, with the following options:
        - `"online"` (default): Enables live syncing with W&B when a network
            connection is available, with real-time updates to visualizations.
        - `"offline"`: Suitable for air-gapped or offline environments; data
            is saved locally and can be synced later. Ensure the run folder
            is preserved to enable future syncing.
        - `"disabled"`: Disables all W&B functionality, making the run’s methods
            no-ops. Typically used in testing to bypass W&B operations.
        - `"shared"`: (This is an experimental feature). Allows multiple processes,
            possibly on different machines, to simultaneously log to the same run.
            In this approach you use a primary node and one or more worker nodes
            to log data to the same run. Within the primary node you
            initialize a run. For each worker node, initialize a run
            using the run ID used by the primary node.
        force: Determines if a W&B login is required to run the script. If `True`,
            the user must be logged in to W&B; otherwise, the script will not
            proceed. If `False` (default), the script can proceed without a login,
            switching to offline mode if the user is not logged in.
        reinit: Shorthand for the "reinit" setting. Determines the behavior of
            `wandb.init()` when a run is active.
        resume: Controls the behavior when resuming a run with the specified `id`.
            Available options are:
        - `"allow"`: If a run with the specified `id` exists, it will resume
            from the last step; otherwise, a new run will be created.
        - `"never"`: If a run with the specified `id` exists, an error will
            be raised. If no such run is found, a new run will be created.
        - `"must"`: If a run with the specified `id` exists, it will resume
            from the last step. If no run is found, an error will be raised.
        - `"auto"`: Automatically resumes the previous run if it crashed on
            this machine; otherwise, starts a new run.
        - `True`: Deprecated. Use `"auto"` instead.
        - `False`: Deprecated. Use the default behavior (leaving `resume`
            unset) to always start a new run.
            If `resume` is set, `fork_from` and `resume_from` cannot be
            used. When `resume` is unset, the system will always start a new run.
        resume_from: Specifies a moment in a previous run to resume a run from,
            using the format `{run_id}?_step={step}`. This allows users to truncate
            the history logged to a run at an intermediate step and resume logging
            from that step. The target run must be in the same project.
            If an `id` argument is also provided, the `resume_from` argument will
            take precedence.
            `resume`, `resume_from` and `fork_from` cannot be used together, only
            one of them can be used at a time.
            Note that this feature is in beta and may change in the future.
        fork_from: Specifies a point in a previous run from which to fork a new
            run, using the format `{id}?_step={step}`. This creates a new run that
            resumes logging from the specified step in the target run’s history.
            The target run must be part of the current project.
            If an `id` argument is also provided, it must be different from the
            `fork_from` argument, an error will be raised if they are the same.
            `resume`, `resume_from` and `fork_from` cannot be used together, only
            one of them can be used at a time.
            Note that this feature is in beta and may change in the future.
        save_code: Enables saving the main script or notebook to W&B, aiding in
            experiment reproducibility and allowing code comparisons across runs in
            the UI. By default, this is disabled, but you can change the default to
            enable on your settings page.
        tensorboard: Deprecated. Use `sync_tensorboard` instead.
        sync_tensorboard: Enables automatic syncing of W&B logs from TensorBoard
            or TensorBoardX, saving relevant event files for viewing in
            the W&B UI.
        monitor_gym: Enables automatic logging of videos of the environment when
            using OpenAI Gym.
        settings: Specifies a dictionary or `wandb.Settings` object with advanced
            settings for the run.

    Returns:
        A `Run` object.

    Raises:
        Error: If some unknown or internal error happened during the run
            initialization.
        AuthenticationError: If the user failed to provide valid credentials.
        CommError: If there was a problem communicating with the WandB server.
        UsageError: If the user provided invalid arguments.
        KeyboardInterrupt: If user interrupts the run.

    Examples:
    `wandb.init()` returns a `Run` object. Use the run object to log data,
    save artifacts, and manage the run lifecycle.

    ```python
    import wandb

    config = {"lr": 0.01, "batch_size": 32}
    with wandb.init(config=config) as run:
        # Log accuracy and loss to the run
        acc = 0.95  # Example accuracy
        loss = 0.05  # Example loss
        run.log({"accuracy": acc, "loss": loss})
    ```
    """
    init_telemetry = telemetry.TelemetryRecord()

    init_settings = Settings()
    if isinstance(settings, dict):
        init_settings = Settings(**settings)
    elif isinstance(settings, Settings):
        init_settings = settings

    # Explicit function arguments take precedence over settings
    if job_type is not None:
        init_settings.run_job_type = job_type
    if dir is not None:
        init_settings.root_dir = dir  # type: ignore
    if project is not None:
        init_settings.project = project
    if entity is not None:
        init_settings.entity = entity
    if reinit is not None:
        init_settings.reinit = reinit
    if tags is not None:
        init_settings.run_tags = tuple(tags)
    if group is not None:
        init_settings.run_group = group
    if name is not None:
        init_settings.run_name = name
    if notes is not None:
        init_settings.run_notes = notes
    if anonymous is not UNSET:
        init_settings.anonymous = anonymous
    if mode is not None:
        init_settings.mode = mode  # type: ignore
    if resume is not None:
        init_settings.resume = resume  # type: ignore
    if force is not None:
        init_settings.force = force
    # TODO: deprecate "tensorboard" in favor of "sync_tensorboard"
    if tensorboard is not None:
        init_settings.sync_tensorboard = tensorboard
    if sync_tensorboard is not None:
        init_settings.sync_tensorboard = sync_tensorboard
    if save_code is not None:
        init_settings.save_code = save_code
    if id is not None:
        init_settings.run_id = id
    if fork_from is not None:
        init_settings.fork_from = fork_from  # type: ignore
    if resume_from is not None:
        init_settings.resume_from = resume_from  # type: ignore

    if config is not None:
        init_telemetry.feature.set_init_config = True

    wl: wandb_setup._WandbSetup | None = None

    try:
        wl = wandb_setup.singleton()

        wi = _WandbInit(wl, init_telemetry)

        wi.maybe_login(init_settings)
        run_settings, show_warnings = wi.make_run_settings(init_settings)

        if isinstance(run_settings.reinit, bool):
            wi.deprecated_features_used.append(
                (
                    Deprecated(run__reinit_bool=True),
                    "Using a boolean value for 'reinit' is deprecated."
                    " Use 'return_previous' or 'finish_previous' instead.",
                )
            )

        if run_settings.run_id is not None:
            init_telemetry.feature.set_init_id = True
        if run_settings.run_name is not None:
            init_telemetry.feature.set_init_name = True
        if run_settings.run_tags is not None:
            init_telemetry.feature.set_init_tags = True
        if run_settings._offline:
            init_telemetry.feature.offline = True
        if run_settings.fork_from is not None:
            init_telemetry.feature.fork_mode = True
        if run_settings.resume_from is not None:
            init_telemetry.feature.rewind_mode = True

        wi.set_run_id(run_settings)
        wi.set_sync_dir_suffix(run_settings)
        run_printer = printer.new_printer(run_settings)
        show_warnings(run_printer)

        with contextlib.ExitStack() as exit_stack:
            exit_stack.enter_context(wb_logging.log_to_run(run_settings.run_id))

            run_config = wi.make_run_config(
                settings=run_settings,
                config=config,
                config_exclude_keys=config_exclude_keys,
                config_include_keys=config_include_keys,
            )

            if run_settings._noop:
                return wi.make_disabled_run(run_config)

            try_create_root_dir(run_settings)
            exit_stack.enter_context(wi.setup_run_log_directory(run_settings))

            if run_settings._jupyter:
                wi.monkeypatch_ipython(run_settings)

            if monitor_gym:
                _monkeypatch_openai_gym()

            if wandb.patched["tensorboard"]:
                # NOTE: The user may have called the patch function directly.
                init_telemetry.feature.tensorboard_patch = True
            if run_settings.sync_tensorboard:
                _monkeypatch_tensorboard()
                init_telemetry.feature.tensorboard_sync = True

            if run_settings.x_server_side_derived_summary:
                init_telemetry.feature.server_side_derived_summary = True

            run = wi.init(run_settings, run_config, run_printer)

            # Set up automatic Weave integration if Weave is installed
            weave.setup(run_settings.entity, run_settings.project)

            return run

    except KeyboardInterrupt as e:
        if wl:
            wl._get_logger().warning("interrupted", exc_info=e)

        raise

    except Exception as e:
        if wl:
            wl._get_logger().exception("error in wandb.init()", exc_info=e)

        get_sentry().reraise(e)



"""Local file storage handler."""

from __future__ import annotations

import os
import shutil
from collections import deque
from pathlib import Path
from typing import TYPE_CHECKING, Callable
from urllib.parse import ParseResult

from wandb.errors.term import termlog
from wandb.sdk.artifacts.artifact_file_cache import get_artifact_file_cache
from wandb.sdk.artifacts.artifact_manifest_entry import ArtifactManifestEntry
from wandb.sdk.artifacts.storage_handler import DEFAULT_MAX_OBJECTS, StorageHandler
from wandb.sdk.lib.hashutil import B64MD5, md5_file_b64, md5_string
from wandb.sdk.lib.paths import FilePathStr, StrPath, URIStr
from wandb.util import local_file_uri_to_path

from ._timing import TimedIf

if TYPE_CHECKING:
    from wandb.sdk.artifacts.artifact import Artifact
    from wandb.sdk.artifacts.artifact_file_cache import ArtifactFileCache


def _md5_content(path: str) -> B64MD5:
    return md5_file_b64(path)


def _md5_path(path: str) -> B64MD5:
    return md5_string(Path(path).resolve().as_uri())


class LocalFileHandler(StorageHandler):
    """Handles file:// references."""

    _scheme: str
    _cache: ArtifactFileCache

    def __init__(self, scheme: str = "file") -> None:
        """Track files or directories on a local filesystem.

        Expand directories to create an entry for each file contained.
        """
        self._scheme = scheme
        self._cache = get_artifact_file_cache()

    def can_handle(self, parsed_url: ParseResult) -> bool:
        return parsed_url.scheme == self._scheme

    def load_path(
        self,
        manifest_entry: ArtifactManifestEntry,
        local: bool = False,
    ) -> URIStr | FilePathStr:
        if (ref_uri := manifest_entry.ref) is None:
            raise ValueError(f"Cannot add path with no ref: {manifest_entry.path}")

        if not os.path.exists(local_path := local_file_uri_to_path(ref_uri)):
            raise ValueError(
                f"Local file reference: Failed to find file at path {local_path!r}"
            )

        expected_digest = manifest_entry.digest

        path, hit, cache_open = self._cache.check_md5_obj_path(
            b64_md5=expected_digest, size=manifest_entry.size or 0
        )
        if hit:
            return path

        if (digest := md5_file_b64(local_path)) != expected_digest:
            raise ValueError(
                f"Local file reference: Digest mismatch for path {local_path!r}: expected {expected_digest!r} but found {digest!r}"
            )

        # Ensure the parent directory exists
        Path(path).parent.mkdir(parents=True, exist_ok=True)

        with cache_open() as f:
            shutil.copy(local_path, f.name)
        return path

    def store_path(
        self,
        artifact: Artifact,
        path: URIStr | FilePathStr,
        name: StrPath | None = None,
        checksum: bool = True,
        max_objects: int | None = None,
    ) -> list[ArtifactManifestEntry]:
        local_path = local_file_uri_to_path(path)
        max_objects = max_objects or DEFAULT_MAX_OBJECTS

        # If checksum=False, the file's hash should only
        # depend on its absolute path/URI, not its contents

        # Closure func for calculating the file hash from its path
        check_md5: Callable[[str], B64MD5] = _md5_content if checksum else _md5_path

        # We have a single file or directory
        # Note, we follow symlinks for files contained within the directory
        if os.path.isdir(local_path):
            entries: deque[ArtifactManifestEntry] = deque()
            with TimedIf(checksum):
                if checksum:
                    termlog(
                        f"Generating checksum for up to {max_objects!r} files in {local_path!r}... ",
                        newline=False,
                    )

                physical_paths = (
                    os.path.join(root, subpath)
                    for root, _, files in os.walk(local_path)
                    for subpath in files
                )
                for i, physical_path in enumerate(physical_paths):
                    if i >= max_objects:
                        raise ValueError(
                            f"Exceeded {max_objects!r} objects tracked, pass max_objects to add_reference"
                        )

                    # TODO(spencerpearson): this is not a "logical path" in the sense that
                    # `LogicalPath` returns a "logical path"; it's a relative path
                    # **on the local filesystem**.
                    file_path = os.path.relpath(physical_path, start=local_path)
                    artifact_path = os.path.join(name or "", file_path)
                    entry = ArtifactManifestEntry(
                        path=artifact_path,
                        ref=os.path.join(path, file_path),
                        size=os.path.getsize(physical_path),
                        digest=check_md5(physical_path),
                    )
                    entries.append(entry)
            return list(entries)

        if os.path.isfile(local_path):
            return [
                ArtifactManifestEntry(
                    path=name or os.path.basename(local_path),
                    ref=path,
                    size=os.path.getsize(local_path),
                    digest=check_md5(local_path),
                )
            ]
        else:
            # TODO: update error message if we don't allow directories.
            raise ValueError(f"Path {path!r} must be a valid file or directory path")



"""Artifact manifest."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Dict

from pydantic import Field
from typing_extensions import Annotated

from wandb.sdk.lib.hashutil import HexMD5

from ._models.base_model import ArtifactsBase

if TYPE_CHECKING:
    from .artifact_manifest_entry import ArtifactManifestEntry
    from .storage_policy import StoragePolicy


class ArtifactManifest(ArtifactsBase, ABC):
    # Note: we can't name this "version" since it conflicts with the prior
    # `version()` classmethod.
    manifest_version: Annotated[Any, Field(repr=False)]
    entries: Dict[str, ArtifactManifestEntry] = Field(default_factory=dict)  # noqa: UP006

    storage_policy: Annotated[StoragePolicy, Field(exclude=True, repr=False)]

    @classmethod
    def version(cls) -> int:
        return cls.model_fields["manifest_version"].default

    @classmethod
    @abstractmethod
    def from_manifest_json(cls, manifest_json: dict[str, Any]) -> ArtifactManifest:
        if (version := manifest_json.get("version")) is None:
            raise ValueError("Invalid manifest format. Must contain version field.")

        for sub in cls.__subclasses__():
            if sub.version() == version:
                return sub.from_manifest_json(manifest_json)
        raise ValueError("Invalid manifest version.")

    def __len__(self) -> int:
        return len(self.entries)

    @abstractmethod
    def to_manifest_json(self) -> dict[str, Any]:
        raise NotImplementedError

    @abstractmethod
    def digest(self) -> HexMD5:
        raise NotImplementedError

    @abstractmethod
    def size(self) -> int:
        raise NotImplementedError

    def add_entry(self, entry: ArtifactManifestEntry, overwrite: bool = False) -> None:
        if (
            (not overwrite)
            and (old_entry := self.entries.get(entry.path))
            and (entry.digest != old_entry.digest)
        ):
            raise ValueError(f"Cannot add the same path twice: {entry.path!r}")
        self.entries[entry.path] = entry

    def remove_entry(self, entry: ArtifactManifestEntry) -> None:
        try:
            del self.entries[entry.path]
        except LookupError:
            raise FileNotFoundError(f"Cannot remove missing entry: {entry.path!r}")

    def get_entry_by_path(self, path: str) -> ArtifactManifestEntry | None:
        return self.entries.get(path)

    def get_entries_in_directory(self, directory: str) -> list[ArtifactManifestEntry]:
        # entry keys (paths) use forward slash even for windows
        dir_prefix = f"{directory}/"
        return [obj for key, obj in self.entries.items() if key.startswith(dir_prefix)]



# Generated by ariadne-codegen
# Source: tools/graphql_codegen/artifacts/

from __future__ import annotations

from typing import Optional

from pydantic import Field
from typing_extensions import Literal

from wandb._pydantic import GQLResult, Typename

from .enums import ArtifactCollectionState


class DeleteArtifactPortfolio(GQLResult):
    result: Optional[DeleteArtifactPortfolioResult]


class DeleteArtifactPortfolioResult(GQLResult):
    artifact_collection: DeleteArtifactPortfolioResultArtifactCollection = Field(
        alias="artifactCollection"
    )


class DeleteArtifactPortfolioResultArtifactCollection(GQLResult):
    typename__: Typename[
        Literal["ArtifactCollection", "ArtifactPortfolio", "ArtifactSequence"]
    ]
    state: ArtifactCollectionState


DeleteArtifactPortfolio.model_rebuild()
DeleteArtifactPortfolioResult.model_rebuild()



from __future__ import annotations

from typing import TYPE_CHECKING

from wandb.plot.custom_chart import plot_table

if TYPE_CHECKING:
    import wandb
    from wandb.plot.custom_chart import CustomChart


def scatter(
    table: wandb.Table,
    x: str,
    y: str,
    title: str = "",
    split_table: bool = False,
) -> CustomChart:
    """Constructs a scatter plot from a wandb.Table of data.

    Args:
        table: The W&B Table containing the data to visualize.
        x: The name of the column used for the x-axis.
        y: The name of the column used for the y-axis.
        title: The title of the scatter chart.
        split_table: Whether the table should be split into a separate section
            in the W&B UI. If `True`, the table will be displayed in a section named
            "Custom Chart Tables". Default is `False`.

    Returns:
        CustomChart: A custom chart object that can be logged to W&B. To log the
            chart, pass it to `wandb.log()`.
    Example:
    ```python
    import math
    import random
    import wandb

    # Simulate temperature variations at different altitudes over time
    data = [
        [i, random.uniform(-10, 20) - 0.005 * i + 5 * math.sin(i / 50)]
        for i in range(300)
    ]

    # Create W&B table with altitude (m) and temperature (°C) columns
    table = wandb.Table(data=data, columns=["altitude (m)", "temperature (°C)"])

    # Initialize W&B run and log the scatter plot
    with wandb.init(project="temperature-altitude-scatter") as run:
        # Create and log the scatter plot
        scatter_plot = wandb.plot.scatter(
            table=table,
            x="altitude (m)",
            y="temperature (°C)",
            title="Altitude vs Temperature",
        )
        run.log({"altitude-temperature-scatter": scatter_plot})
    ```
    """
    return plot_table(
        data_table=table,
        vega_spec_name="wandb/scatter/v0",
        fields={"x": x, "y": y},
        string_fields={"title": title},
        split_table=split_table,
    )



# Generated by ariadne-codegen
# Source: tools/graphql_codegen/artifacts/

from __future__ import annotations

from typing import Optional

from pydantic import Field

from wandb._pydantic import GQLResult

from .fragments import TypeInfoFragment


class TypeInfo(GQLResult):
    type: Optional[TypeInfoFragment] = Field(alias="__type")


TypeInfo.model_rebuild()



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Final, Literal, Optional, Union, overload

from pydantic import (
    Field,
    PositiveFloat,
    PositiveInt,
    StrictFloat,
    StrictInt,
    field_validator,
)
from typing_extensions import Annotated, TypeAlias, override

from wandb._pydantic import GQLBase
from wandb.automations._validators import LenientStrEnum

from .expressions import FilterExpr
from .operators import BaseOp, RichReprResult

if TYPE_CHECKING:
    from wandb.automations.events import RunMetricFilter

# Maps MongoDB comparison operators -> Python literal (str) representations
MONGO2PY_OPS: Final[dict[str, str]] = {
    "$eq": "==",
    "$ne": "!=",
    "$gt": ">",
    "$lt": "<",
    "$gte": ">=",
    "$lte": "<=",
}
# Reverse mapping from Python literal (str) -> MongoDB operator key
PY2MONGO_OPS: Final[dict[str, str]] = {v: k for k, v in MONGO2PY_OPS.items()}

# Type hint for positive numbers (int or float)
PosNum: TypeAlias = Union[PositiveInt, PositiveFloat]


class Agg(LenientStrEnum):  # from: Aggregation
    """Supported run metric aggregation operations."""

    MAX = "MAX"
    MIN = "MIN"
    AVERAGE = "AVERAGE"

    # Shorter aliases for convenience
    AVG = AVERAGE


class ChangeType(LenientStrEnum):  # from: RunMetricChangeType
    """Describes the type of metric change as absolute or relative.

    ABSOLUTE: The arithmetic difference between the current vs. prior values.
    RELATIVE: The percentage change between the current vs. prior values.
    """

    ABSOLUTE = "ABSOLUTE"
    RELATIVE = "RELATIVE"

    # Shorter aliases for convenience
    ABS = ABSOLUTE
    REL = RELATIVE


class ChangeDir(LenientStrEnum):  # from: RunMetricChangeDirection
    """Describes the direction of the metric change."""

    INCREASE = "INCREASE"
    DECREASE = "DECREASE"
    ANY = "ANY"

    # Shorter aliases for convenience
    INC = INCREASE
    DEC = DECREASE


class BaseMetricFilter(GQLBase, ABC, extra="forbid"):
    name: str
    """Name of the observed metric."""

    agg: Optional[Agg]
    """Aggregate operation, if any, to apply over the window size."""

    window: PositiveInt
    """Size of the metric aggregation window (ignored if `agg` is ``None``)."""

    # ------------------------------------------------------------------------------
    cmp: Optional[str]
    """Comparison operator between the metric expression (left) vs. the threshold or target (right)."""  # noqa: W505

    # ------------------------------------------------------------------------------
    threshold: Union[StrictInt, StrictFloat]
    """Threshold value to compare against."""

    def __and__(self, other: Any) -> RunMetricFilter:
        """Returns `(metric_filter & run_filter)` as a `RunMetricFilter`."""
        from wandb.automations.events import RunMetricFilter

        if isinstance(run_filter := other, (BaseOp, FilterExpr)):
            # Treat `other` as a run filter and build a RunMetricEvent. Let the
            # metric filter validators wrap or nest as appropriate.
            return RunMetricFilter(run=run_filter, metric=self)
        return NotImplemented

    def __rand__(self, other: BaseOp | FilterExpr) -> RunMetricFilter:
        """Ensures `&` is commutative for run and metric filters.

        I.e. `(run_filter & metric_filter) == (metric_filter & run_filter)`.
        """
        return self.__and__(other)

    @abstractmethod
    def __repr__(self) -> str:
        """Returns the text representation of the metric filter."""
        raise NotImplementedError

    @override
    def __rich_repr__(self) -> RichReprResult:
        """Returns the `rich` pretty-print representation of the metric filter."""
        # See: https://rich.readthedocs.io/en/stable/pretty.html#rich-repr-protocol
        yield None, repr(self)


class MetricThresholdFilter(BaseMetricFilter):  # from: RunMetricThresholdFilter
    """Filter that compares an **absolute** metric value against a user-defined threshold.

    The value may be a single value or an aggregated result over a window of
    multiple values.
    """

    name: str
    agg: Annotated[Optional[Agg], Field(alias="agg_op")] = None
    window: Annotated[PositiveInt, Field(alias="window_size")] = 1

    cmp: Annotated[Literal["$gte", "$gt", "$lt", "$lte"], Field(alias="cmp_op")]
    """Comparison operator between the metric value (left) vs. the threshold (right)."""

    threshold: Union[StrictInt, StrictFloat]

    @field_validator("cmp", mode="before")
    def _validate_cmp(cls, v: Any) -> Any:
        # Be helpful: e.g. ">" -> "$gt"
        return PY2MONGO_OPS.get(v.strip(), v) if isinstance(v, str) else v

    def __repr__(self) -> str:
        metric = f"{self.agg.value}({self.name})" if self.agg else self.name
        op = MONGO2PY_OPS.get(self.cmp, self.cmp)
        return repr(rf"{metric} {op} {self.threshold}")


class MetricChangeFilter(BaseMetricFilter):  # from: RunMetricChangeFilter
    """Filter that compares a **change** in a metric value to a user-defined threshold.

    The change is calculated over "tumbling" windows, i.e. the difference
    between the current window and the non-overlapping prior window.
    """

    name: str
    agg: Annotated[Optional[Agg], Field(alias="agg_op")] = None
    window: Annotated[PositiveInt, Field(alias="current_window_size")] = 1

    # `prior_window` is only for `RUN_METRIC_CHANGE` events
    prior_window: Annotated[
        PositiveInt,
        # By default, set `window -> prior_window` if the latter wasn't provided.
        Field(alias="prior_window_size", default_factory=lambda data: data["window"]),
    ]
    """Size of the "prior" metric aggregation window (ignored if `agg` is ``None``).

    If omitted, defaults to the size of the current window.
    """

    # ------------------------------------------------------------------------------
    # NOTE:
    # - The "comparison" operator isn't actually part of the backend schema,
    #   but it's defined here for consistency -- and ignored otherwise.
    # - In the backend, it's effectively "$gte" or "$lte", depending on the sign
    #   (change_dir), though again, this is not explicit in the schema.
    cmp: Annotated[None, Field(frozen=True, exclude=True, repr=False)] = None
    """Ignored."""

    # ------------------------------------------------------------------------------
    change_type: ChangeType
    change_dir: ChangeDir
    threshold: Annotated[PosNum, Field(alias="change_amount")]

    def __repr__(self) -> str:
        metric = f"{self.agg.value}({self.name})" if self.agg else self.name
        verb = (
            "changes"
            if (self.change_dir is ChangeDir.ANY)
            else f"{self.change_dir.value.lower()}s"
        )

        fmt_spec = ".2%" if (self.change_type is ChangeType.REL) else ""
        amt = f"{self.threshold:{fmt_spec}}"
        return repr(rf"{metric} {verb} {amt}")


class BaseMetricOperand(GQLBase, ABC, extra="forbid"):
    def gt(self, value: int | float, /) -> MetricThresholdFilter:
        """Returns a filter that watches for `metric_expr > threshold`."""
        return self > value

    def lt(self, value: int | float, /) -> MetricThresholdFilter:
        """Returns a filter that watches for `metric_expr < threshold`."""
        return self < value

    def gte(self, value: int | float, /) -> MetricThresholdFilter:
        """Returns a filter that watches for `metric_expr >= threshold`."""
        return self >= value

    def lte(self, value: int | float, /) -> MetricThresholdFilter:
        """Returns a filter that watches for `metric_expr <= threshold`."""
        return self <= value

    # Overloads to implement:
    # - `(metric_operand > threshold) -> MetricThresholdFilter`
    # - `(metric_operand < threshold) -> MetricThresholdFilter`
    # - `(metric_operand >= threshold) -> MetricThresholdFilter`
    # - `(metric_operand <= threshold) -> MetricThresholdFilter`
    def __gt__(self, other: Any) -> MetricThresholdFilter:
        if isinstance(other, (int, float)):
            return MetricThresholdFilter(**dict(self), cmp="$gt", threshold=other)
        return NotImplemented

    def __lt__(self, other: Any) -> MetricThresholdFilter:
        if isinstance(other, (int, float)):
            return MetricThresholdFilter(**dict(self), cmp="$lt", threshold=other)
        return NotImplemented

    def __ge__(self, other: Any) -> MetricThresholdFilter:
        if isinstance(other, (int, float)):
            return MetricThresholdFilter(**dict(self), cmp="$gte", threshold=other)
        return NotImplemented

    def __le__(self, other: Any) -> MetricThresholdFilter:
        if isinstance(other, (int, float)):
            return MetricThresholdFilter(**dict(self), cmp="$lte", threshold=other)
        return NotImplemented

    @overload
    def changes_by(self, *, diff: PosNum, frac: None) -> MetricChangeFilter: ...

    @overload
    def changes_by(self, *, diff: None, frac: PosNum) -> MetricChangeFilter: ...

    @overload  # NOTE: This overload is for internal use only.
    def changes_by(
        self, *, diff: PosNum | None, frac: PosNum | None, _dir: ChangeDir
    ) -> MetricChangeFilter: ...

    def changes_by(
        self,
        *,
        diff: PosNum | None = None,
        frac: PosNum | None = None,
        _dir: ChangeDir = ChangeDir.ANY,
    ) -> MetricChangeFilter:
        """Returns a filter that watches for a numerical increase OR decrease in a metric.

        Exactly one of `frac` or `diff` must be provided.

        Args:
            diff: If given, arithmetic difference that must be observed in the metric.
                Must be positive.
            frac: If given, fractional (relative) change that must be observed in the
                metric. Must be positive. For example, `frac=0.1` denotes a 10% relative
                increase or decrease.
        """
        if (
            # Enforce mutually exclusive keyword args
            ((frac is None) and (diff is None))
            or ((frac is not None) and (diff is not None))
        ):
            raise ValueError("Must provide exactly one of `frac` or `diff`")

        # Enforce positive values
        if (frac is not None) and (frac <= 0):
            raise ValueError(f"Expected positive threshold, got: {frac=}")
        if (diff is not None) and (diff <= 0):
            raise ValueError(f"Expected positive threshold, got: {diff=}")

        if diff is None:
            kws = dict(change_dir=_dir, change_type=ChangeType.REL, threshold=frac)
        else:
            kws = dict(change_dir=_dir, change_type=ChangeType.ABS, threshold=diff)
        return MetricChangeFilter(**dict(self), **kws)

    @overload
    def increases_by(self, *, diff: PosNum, frac: None) -> MetricChangeFilter: ...

    @overload
    def increases_by(self, *, diff: None, frac: PosNum) -> MetricChangeFilter: ...

    def increases_by(
        self, *, diff: PosNum | None = None, frac: PosNum | None = None
    ) -> MetricChangeFilter:
        """Returns a filter that watches for a numerical increase in a metric.

        Arguments mirror those of `.changes_by()`.
        """
        return self.changes_by(diff=diff, frac=frac, _dir=ChangeDir.INC)

    @overload
    def decreases_by(self, *, diff: PosNum, frac: None) -> MetricChangeFilter: ...

    @overload
    def decreases_by(self, *, diff: None, frac: PosNum) -> MetricChangeFilter: ...

    def decreases_by(
        self, *, diff: PosNum | None = None, frac: PosNum | None = None
    ) -> MetricChangeFilter:
        """Returns a filter that watches for a numerical decrease in a metric.

        Arguments mirror those of `.changes_by()`.
        """
        return self.changes_by(diff=diff, frac=frac, _dir=ChangeDir.DEC)


class MetricVal(BaseMetricOperand):
    """Represents a single metric value when defining metric event filters."""

    name: str

    # Allow conversion of a single-value metric into an aggregated expression.
    def max(self, window: int) -> MetricAgg:
        return MetricAgg(name=self.name, agg=Agg.MAX, window=window)

    def min(self, window: int) -> MetricAgg:
        return MetricAgg(name=self.name, agg=Agg.MIN, window=window)

    def avg(self, window: int) -> MetricAgg:
        return MetricAgg(name=self.name, agg=Agg.AVG, window=window)

    # Aliased method for users familiar with e.g. torch/tf/numpy/pandas/polars/etc.
    def mean(self, window: int) -> MetricAgg:
        return self.avg(window=window)

    def zscore(self, window: int) -> ZScoreMetricOperand:
        """Returns a z-score metric builder for fluent filter construction.

        Use with comparison operators to create z-score filters:
        - `metric.zscore(30) > 3` - detects z-score increases above 3 std devs
        - `metric.zscore(30) < -3` - detects z-score decreases below -3 std devs
        - `metric.zscore(30).abs() > 3` - detects abs z-score deviations above 3 std devs

        Note:
        - The `>=` operator behaves the same as `>`, and `<=` behaves the same as `<`.
        """
        return ZScoreMetricOperand(name=self.name, window=window)


class MetricAgg(BaseMetricOperand):
    """Represents an aggregated metric value when defining metric event filters."""

    name: str
    agg: Annotated[Agg, Field(alias="agg_op")]
    window: Annotated[PositiveInt, Field(alias="window_size")]


class ZScoreMetricOperand(GQLBase, extra="forbid"):
    """Helper class to build z-score metric filters with comparison operators.

    This class enables fluent construction of z-score filters using Python
    comparison operators (>, <, >=, <=) and the builtin abs() function.

    Note: When defining a z-score threshold, the `>` and `>=` operators are
    interchangeable, as are the `<=` and `<` operators, since the z-score defines
    a threshold on a continuous value. At runtime, the filter is evaluated
    using the inclusive operators (`>=` or `<=`).
    """

    name: str
    """Name of the metric to monitor."""

    window: PositiveInt
    """Size of the window to calculate the metric mean and standard deviation over."""

    is_absolute: bool = Field(default=False, repr=False)
    """Whether to check the absolute value of the z-score (ignoring direction)."""

    def lt(self, value: int | float, /) -> MetricZScoreFilter:
        """Returns a filter that watches for `zscore(metric) < -threshold`.

        Args:
            value: The z-score threshold value to compare against.
                   The absolute value is used as the threshold.
        """
        if self.is_absolute:
            raise ValueError("Cannot use absolute z-score with < operator")

        if value >= 0:
            raise ValueError("Negative z-score threshold required")

        return MetricZScoreFilter(
            name=self.name,
            window=self.window,
            change_dir=ChangeDir.DECREASE,
            threshold=abs(value),
        )

    def __lt__(self, value: int | float, /) -> MetricZScoreFilter:
        return self.lt(value)

    def __le__(self, value: int | float, /) -> MetricZScoreFilter:
        """Alias for `<` operator - behaves identically to `__lt__`.

        Returns a filter that watches for `zscore(metric) < -threshold`.
        Note: `<=` and `<` are treated as equivalent for z-score filters.
        """
        return self.lt(value)

    def gt(self, value: int | float, /) -> MetricZScoreFilter:
        """Returns a filter that watches for `zscore(metric) > threshold`.

        If `is_absolute` is True, watches for `abs(zscore(metric)) > threshold`.

        Args:
            value: The z-score threshold value to compare against.
                   The absolute value is used as the threshold.
        """
        if value <= 0:
            raise ValueError(f"Expected positive threshold, got: {value=}")

        return MetricZScoreFilter(
            name=self.name,
            window=self.window,
            change_dir=ChangeDir.ANY if self.is_absolute else ChangeDir.INCREASE,
            threshold=abs(value),
        )

    def __gt__(self, value: int | float, /) -> MetricZScoreFilter:
        return self.gt(value)

    def __ge__(self, value: int | float, /) -> MetricZScoreFilter:
        """Alias for `>` operator - behaves identically to `__gt__`.

        Returns a filter that watches for `zscore(metric) > threshold`.
        If `is_absolute` is True, watches for `abs(zscore(metric)) > threshold`.
        Note: `>=` and `>` are treated as equivalent for z-score filters.
        """
        return self.gt(value)

    def __abs__(self) -> ZScoreMetricOperand:
        """Returns a z-score filter that checks the absolute value.

        This allows watching for z-score deviations in either direction.
        Use with comparison operators: `abs(metric.zscore(window)) > threshold`.
        """
        return self.model_copy(update={"is_absolute": True})

    def abs(self) -> ZScoreMetricOperand:
        """Returns a z-score filter that checks the absolute value.

        Alias for `__abs__()` that can be called as a method.
        Allows using either `abs(zscore)` or `zscore.abs()`.
        """
        return self.__abs__()


class MetricZScoreFilter(GQLBase, extra="forbid"):
    name: str
    """Name of the observed metric."""

    window: Annotated[PositiveInt, Field(alias="window_size")] = 30
    """Size of the window to calculate the metric mean and standard deviation over."""

    threshold: PosNum = 3.0
    """Threshold for the z-score."""

    change_dir: ChangeDir = ChangeDir.ANY
    """Direction of the z-score change to watch for."""

    def __and__(self, other: Any) -> RunMetricFilter:
        """Returns `(metric_filter & run_filter)` as a `RunMetricFilter`."""
        from wandb.automations.events import RunMetricFilter

        if isinstance(run_filter := other, (BaseOp, FilterExpr)):
            # Treat `other` as a run filter and build a RunMetricEvent. Let the
            # metric filter validators wrap or nest as appropriate.
            return RunMetricFilter(run=run_filter, metric=self)
        return NotImplemented

    def __rand__(self, other: BaseOp | FilterExpr) -> RunMetricFilter:
        """Ensures `&` is commutative for run and metric filters.

        I.e. `(run_filter & metric_filter) == (metric_filter & run_filter)`.
        """
        return self.__and__(other)

    def __repr__(self) -> str:
        if self.change_dir is ChangeDir.ANY:
            return repr(rf"abs(zscore({self.name!r})) > {self.threshold}")
        elif self.change_dir is ChangeDir.DECREASE:
            return repr(rf"zscore({self.name!r}) < -{self.threshold}")
        else:  # ChangeDir.INCREASE
            return repr(rf"zscore({self.name!r}) > +{self.threshold}")

    @override
    def __rich_repr__(self) -> RichReprResult:
        """Returns the `rich` pretty-print representation of the metric filter."""
        # See: https://rich.readthedocs.io/en/stable/pretty.html#rich-repr-protocol
        yield None, repr(self)



from __future__ import annotations

import logging
from multiprocessing.process import BaseProcess
from typing import TYPE_CHECKING

from typing_extensions import override

from .interface_shared import InterfaceShared

if TYPE_CHECKING:
    from queue import Queue

    from wandb.proto import wandb_internal_pb2 as pb
    from wandb.sdk.mailbox.mailbox_handle import MailboxHandle


logger = logging.getLogger("wandb")


class InterfaceQueue(InterfaceShared):
    """Legacy implementation of InterfaceShared.

    This was used by legacy-service to pass messages back to itself before
    the existence of wandb-core. It may be removed once legacy-service is
    completely removed (including its use in `wandb sync`).

    Since it was used by the internal service, it does not implement
    the "deliver" methods, which are only used in the client.
    """

    def __init__(
        self,
        record_q: Queue[pb.Record] | None = None,
        result_q: Queue[pb.Result] | None = None,
        process: BaseProcess | None = None,
    ) -> None:
        self.record_q = record_q
        self.result_q = result_q
        self._process = process
        super().__init__()

    @override
    def _publish(self, record: pb.Record, *, nowait: bool = False) -> None:
        if self._process and not self._process.is_alive():
            raise Exception("The wandb backend process has shutdown")
        if self.record_q:
            self.record_q.put(record)

    @override
    async def deliver_async(
        self,
        record: pb.Record,
    ) -> MailboxHandle[pb.Result]:
        raise NotImplementedError

    @override
    def _deliver(self, record: pb.Record) -> MailboxHandle[pb.Result]:
        raise NotImplementedError



# Generated by ariadne-codegen
# Source: tools/graphql_codegen/artifacts/

from __future__ import annotations

from typing import List, Optional

from pydantic import Field

from wandb._pydantic import GQLResult

from .fragments import ArtifactFragment, PageInfoFragment


class RunInputArtifacts(GQLResult):
    project: Optional[RunInputArtifactsProject]


class RunInputArtifactsProject(GQLResult):
    run: Optional[RunInputArtifactsProjectRun]


class RunInputArtifactsProjectRun(GQLResult):
    artifacts: Optional[RunInputArtifactsProjectRunArtifacts]


class RunInputArtifactsProjectRunArtifacts(GQLResult):
    total_count: int = Field(alias="totalCount")
    page_info: PageInfoFragment = Field(alias="pageInfo")
    edges: List[RunInputArtifactsProjectRunArtifactsEdges]


class RunInputArtifactsProjectRunArtifactsEdges(GQLResult):
    node: Optional[ArtifactFragment]


RunInputArtifacts.model_rebuild()
RunInputArtifactsProject.model_rebuild()
RunInputArtifactsProjectRun.model_rebuild()
RunInputArtifactsProjectRunArtifacts.model_rebuild()
RunInputArtifactsProjectRunArtifactsEdges.model_rebuild()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: wandb/proto/wandb_api.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from wandb.proto import wandb_settings_pb2 as wandb_dot_proto_dot_wandb__settings__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1bwandb/proto/wandb_api.proto\x12\x0ewandb_internal\x1a wandb/proto/wandb_settings.proto\"B\n\x14ServerApiInitRequest\x12*\n\x08settings\x18\x01 \x01(\x0b\x32\x18.wandb_internal.Settings\"b\n\nApiRequest\x12I\n\x18read_run_history_request\x18\x01 \x01(\x0b\x32%.wandb_internal.ReadRunHistoryRequestH\x00\x42\t\n\x07request\".\n\x15ServerApiInitResponse\x12\x15\n\rerror_message\x18\x01 \x01(\t\"\xa6\x01\n\x0b\x41piResponse\x12K\n\x19read_run_history_response\x18\x01 \x01(\x0b\x32&.wandb_internal.ReadRunHistoryResponseH\x00\x12>\n\x12\x61pi_error_response\x18\x02 \x01(\x0b\x32 .wandb_internal.ApiErrorResponseH\x00\x42\n\n\x08response\"#\n\x10\x41piErrorResponse\x12\x0f\n\x07message\x18\x01 \x01(\t\"\xee\x01\n\x15ReadRunHistoryRequest\x12\x43\n\x15scan_run_history_init\x18\x01 \x01(\x0b\x32\".wandb_internal.ScanRunHistoryInitH\x00\x12:\n\x10scan_run_history\x18\x02 \x01(\x0b\x32\x1e.wandb_internal.ScanRunHistoryH\x00\x12I\n\x18scan_run_history_cleanup\x18\x03 \x01(\x0b\x32%.wandb_internal.ScanRunHistoryCleanupH\x00\x42\t\n\x07request\"\xff\x01\n\x16ReadRunHistoryResponse\x12K\n\x15scan_run_history_init\x18\x01 \x01(\x0b\x32*.wandb_internal.ScanRunHistoryInitResponseH\x00\x12\x39\n\x0brun_history\x18\x02 \x01(\x0b\x32\".wandb_internal.RunHistoryResponseH\x00\x12Q\n\x18scan_run_history_cleanup\x18\x03 \x01(\x0b\x32-.wandb_internal.ScanRunHistoryCleanupResponseH\x00\x42\n\n\x08response\"f\n\x12ScanRunHistoryInit\x12\x0e\n\x06\x65ntity\x18\x01 \x01(\t\x12\x0f\n\x07project\x18\x02 \x01(\t\x12\x0e\n\x06run_id\x18\x03 \x01(\t\x12\x0c\n\x04keys\x18\x04 \x03(\t\x12\x11\n\tuse_cache\x18\x05 \x01(\x08\"0\n\x1aScanRunHistoryInitResponse\x12\x12\n\nrequest_id\x18\x01 \x01(\x05\"H\n\x0eScanRunHistory\x12\x10\n\x08min_step\x18\x01 \x01(\x03\x12\x10\n\x08max_step\x18\x02 \x01(\x03\x12\x12\n\nrequest_id\x18\x03 \x01(\x05\"F\n\x12RunHistoryResponse\x12\x30\n\x0chistory_rows\x18\x01 \x03(\x0b\x32\x1a.wandb_internal.HistoryRow\"G\n\nHistoryRow\x12\x39\n\rhistory_items\x18\x01 \x03(\x0b\x32\".wandb_internal.ParquetHistoryItem\"5\n\x12ParquetHistoryItem\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x12\n\nvalue_json\x18\x10 \x01(\t\"+\n\x15ScanRunHistoryCleanup\x12\x12\n\nrequest_id\x18\x01 \x01(\x05\"\x1f\n\x1dScanRunHistoryCleanupResponseB\x1bZ\x19\x63ore/pkg/service_go_protob\x06proto3')



_SERVERAPIINITREQUEST = DESCRIPTOR.message_types_by_name['ServerApiInitRequest']
_APIREQUEST = DESCRIPTOR.message_types_by_name['ApiRequest']
_SERVERAPIINITRESPONSE = DESCRIPTOR.message_types_by_name['ServerApiInitResponse']
_APIRESPONSE = DESCRIPTOR.message_types_by_name['ApiResponse']
_APIERRORRESPONSE = DESCRIPTOR.message_types_by_name['ApiErrorResponse']
_READRUNHISTORYREQUEST = DESCRIPTOR.message_types_by_name['ReadRunHistoryRequest']
_READRUNHISTORYRESPONSE = DESCRIPTOR.message_types_by_name['ReadRunHistoryResponse']
_SCANRUNHISTORYINIT = DESCRIPTOR.message_types_by_name['ScanRunHistoryInit']
_SCANRUNHISTORYINITRESPONSE = DESCRIPTOR.message_types_by_name['ScanRunHistoryInitResponse']
_SCANRUNHISTORY = DESCRIPTOR.message_types_by_name['ScanRunHistory']
_RUNHISTORYRESPONSE = DESCRIPTOR.message_types_by_name['RunHistoryResponse']
_HISTORYROW = DESCRIPTOR.message_types_by_name['HistoryRow']
_PARQUETHISTORYITEM = DESCRIPTOR.message_types_by_name['ParquetHistoryItem']
_SCANRUNHISTORYCLEANUP = DESCRIPTOR.message_types_by_name['ScanRunHistoryCleanup']
_SCANRUNHISTORYCLEANUPRESPONSE = DESCRIPTOR.message_types_by_name['ScanRunHistoryCleanupResponse']
ServerApiInitRequest = _reflection.GeneratedProtocolMessageType('ServerApiInitRequest', (_message.Message,), {
  'DESCRIPTOR' : _SERVERAPIINITREQUEST,
  '__module__' : 'wandb.proto.wandb_api_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.ServerApiInitRequest)
  })
_sym_db.RegisterMessage(ServerApiInitRequest)

ApiRequest = _reflection.GeneratedProtocolMessageType('ApiRequest', (_message.Message,), {
  'DESCRIPTOR' : _APIREQUEST,
  '__module__' : 'wandb.proto.wandb_api_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.ApiRequest)
  })
_sym_db.RegisterMessage(ApiRequest)

ServerApiInitResponse = _reflection.GeneratedProtocolMessageType('ServerApiInitResponse', (_message.Message,), {
  'DESCRIPTOR' : _SERVERAPIINITRESPONSE,
  '__module__' : 'wandb.proto.wandb_api_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.ServerApiInitResponse)
  })
_sym_db.RegisterMessage(ServerApiInitResponse)

ApiResponse = _reflection.GeneratedProtocolMessageType('ApiResponse', (_message.Message,), {
  'DESCRIPTOR' : _APIRESPONSE,
  '__module__' : 'wandb.proto.wandb_api_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.ApiResponse)
  })
_sym_db.RegisterMessage(ApiResponse)

ApiErrorResponse = _reflection.GeneratedProtocolMessageType('ApiErrorResponse', (_message.Message,), {
  'DESCRIPTOR' : _APIERRORRESPONSE,
  '__module__' : 'wandb.proto.wandb_api_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.ApiErrorResponse)
  })
_sym_db.RegisterMessage(ApiErrorResponse)

ReadRunHistoryRequest = _reflection.GeneratedProtocolMessageType('ReadRunHistoryRequest', (_message.Message,), {
  'DESCRIPTOR' : _READRUNHISTORYREQUEST,
  '__module__' : 'wandb.proto.wandb_api_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.ReadRunHistoryRequest)
  })
_sym_db.RegisterMessage(ReadRunHistoryRequest)

ReadRunHistoryResponse = _reflection.GeneratedProtocolMessageType('ReadRunHistoryResponse', (_message.Message,), {
  'DESCRIPTOR' : _READRUNHISTORYRESPONSE,
  '__module__' : 'wandb.proto.wandb_api_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.ReadRunHistoryResponse)
  })
_sym_db.RegisterMessage(ReadRunHistoryResponse)

ScanRunHistoryInit = _reflection.GeneratedProtocolMessageType('ScanRunHistoryInit', (_message.Message,), {
  'DESCRIPTOR' : _SCANRUNHISTORYINIT,
  '__module__' : 'wandb.proto.wandb_api_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.ScanRunHistoryInit)
  })
_sym_db.RegisterMessage(ScanRunHistoryInit)

ScanRunHistoryInitResponse = _reflection.GeneratedProtocolMessageType('ScanRunHistoryInitResponse', (_message.Message,), {
  'DESCRIPTOR' : _SCANRUNHISTORYINITRESPONSE,
  '__module__' : 'wandb.proto.wandb_api_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.ScanRunHistoryInitResponse)
  })
_sym_db.RegisterMessage(ScanRunHistoryInitResponse)

ScanRunHistory = _reflection.GeneratedProtocolMessageType('ScanRunHistory', (_message.Message,), {
  'DESCRIPTOR' : _SCANRUNHISTORY,
  '__module__' : 'wandb.proto.wandb_api_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.ScanRunHistory)
  })
_sym_db.RegisterMessage(ScanRunHistory)

RunHistoryResponse = _reflection.GeneratedProtocolMessageType('RunHistoryResponse', (_message.Message,), {
  'DESCRIPTOR' : _RUNHISTORYRESPONSE,
  '__module__' : 'wandb.proto.wandb_api_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.RunHistoryResponse)
  })
_sym_db.RegisterMessage(RunHistoryResponse)

HistoryRow = _reflection.GeneratedProtocolMessageType('HistoryRow', (_message.Message,), {
  'DESCRIPTOR' : _HISTORYROW,
  '__module__' : 'wandb.proto.wandb_api_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.HistoryRow)
  })
_sym_db.RegisterMessage(HistoryRow)

ParquetHistoryItem = _reflection.GeneratedProtocolMessageType('ParquetHistoryItem', (_message.Message,), {
  'DESCRIPTOR' : _PARQUETHISTORYITEM,
  '__module__' : 'wandb.proto.wandb_api_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.ParquetHistoryItem)
  })
_sym_db.RegisterMessage(ParquetHistoryItem)

ScanRunHistoryCleanup = _reflection.GeneratedProtocolMessageType('ScanRunHistoryCleanup', (_message.Message,), {
  'DESCRIPTOR' : _SCANRUNHISTORYCLEANUP,
  '__module__' : 'wandb.proto.wandb_api_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.ScanRunHistoryCleanup)
  })
_sym_db.RegisterMessage(ScanRunHistoryCleanup)

ScanRunHistoryCleanupResponse = _reflection.GeneratedProtocolMessageType('ScanRunHistoryCleanupResponse', (_message.Message,), {
  'DESCRIPTOR' : _SCANRUNHISTORYCLEANUPRESPONSE,
  '__module__' : 'wandb.proto.wandb_api_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.ScanRunHistoryCleanupResponse)
  })
_sym_db.RegisterMessage(ScanRunHistoryCleanupResponse)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z\031core/pkg/service_go_proto'
  _SERVERAPIINITREQUEST._serialized_start=81
  _SERVERAPIINITREQUEST._serialized_end=147
  _APIREQUEST._serialized_start=149
  _APIREQUEST._serialized_end=247
  _SERVERAPIINITRESPONSE._serialized_start=249
  _SERVERAPIINITRESPONSE._serialized_end=295
  _APIRESPONSE._serialized_start=298
  _APIRESPONSE._serialized_end=464
  _APIERRORRESPONSE._serialized_start=466
  _APIERRORRESPONSE._serialized_end=501
  _READRUNHISTORYREQUEST._serialized_start=504
  _READRUNHISTORYREQUEST._serialized_end=742
  _READRUNHISTORYRESPONSE._serialized_start=745
  _READRUNHISTORYRESPONSE._serialized_end=1000
  _SCANRUNHISTORYINIT._serialized_start=1002
  _SCANRUNHISTORYINIT._serialized_end=1104
  _SCANRUNHISTORYINITRESPONSE._serialized_start=1106
  _SCANRUNHISTORYINITRESPONSE._serialized_end=1154
  _SCANRUNHISTORY._serialized_start=1156
  _SCANRUNHISTORY._serialized_end=1228
  _RUNHISTORYRESPONSE._serialized_start=1230
  _RUNHISTORYRESPONSE._serialized_end=1300
  _HISTORYROW._serialized_start=1302
  _HISTORYROW._serialized_end=1373
  _PARQUETHISTORYITEM._serialized_start=1375
  _PARQUETHISTORYITEM._serialized_end=1428
  _SCANRUNHISTORYCLEANUP._serialized_start=1430
  _SCANRUNHISTORYCLEANUP._serialized_end=1473
  _SCANRUNHISTORYCLEANUPRESPONSE._serialized_start=1475
  _SCANRUNHISTORYCLEANUPRESPONSE._serialized_end=1506
# @@protoc_insertion_point(module_scope)



"""Storage layout."""

from __future__ import annotations

from enum import Enum


class StorageLayout(str, Enum):
    V1 = "V1"
    V2 = "V2"

    @classmethod
    def from_env(cls) -> StorageLayout:
        from wandb.env import get_use_v1_artifacts

        return cls.V1 if get_use_v1_artifacts() else cls.V2



# Generated by ariadne-codegen
# Source: tools/graphql_codegen/artifacts/

from __future__ import annotations

from typing import List, Optional

from pydantic import Field
from typing_extensions import Literal

from wandb._pydantic import GQLResult, Typename

from .fragments import FileFragment, PageInfoFragment


class GetArtifactMembershipFiles(GQLResult):
    project: Optional[GetArtifactMembershipFilesProject]


class GetArtifactMembershipFilesProject(GQLResult):
    artifact_collection: Optional[
        GetArtifactMembershipFilesProjectArtifactCollection
    ] = Field(alias="artifactCollection")


class GetArtifactMembershipFilesProjectArtifactCollection(GQLResult):
    typename__: Typename[
        Literal["ArtifactCollection", "ArtifactPortfolio", "ArtifactSequence"]
    ]
    artifact_membership: Optional[
        GetArtifactMembershipFilesProjectArtifactCollectionArtifactMembership
    ] = Field(alias="artifactMembership")


class GetArtifactMembershipFilesProjectArtifactCollectionArtifactMembership(GQLResult):
    files: Optional[
        GetArtifactMembershipFilesProjectArtifactCollectionArtifactMembershipFiles
    ]


class GetArtifactMembershipFilesProjectArtifactCollectionArtifactMembershipFiles(
    GQLResult
):
    total_count: Optional[int] = Field(alias="totalCount", default=None)
    page_info: PageInfoFragment = Field(alias="pageInfo")
    edges: List[
        GetArtifactMembershipFilesProjectArtifactCollectionArtifactMembershipFilesEdges
    ]


class GetArtifactMembershipFilesProjectArtifactCollectionArtifactMembershipFilesEdges(
    GQLResult
):
    node: Optional[FileFragment]


GetArtifactMembershipFiles.model_rebuild()
GetArtifactMembershipFilesProject.model_rebuild()
GetArtifactMembershipFilesProjectArtifactCollection.model_rebuild()
GetArtifactMembershipFilesProjectArtifactCollectionArtifactMembership.model_rebuild()
GetArtifactMembershipFilesProjectArtifactCollectionArtifactMembershipFiles.model_rebuild()
GetArtifactMembershipFilesProjectArtifactCollectionArtifactMembershipFilesEdges.model_rebuild()



from __future__ import annotations

from enum import Enum
from functools import lru_cache, partial
from typing import TYPE_CHECKING, Any, Collection

from wandb_gql import gql

from wandb._strutils import ensureprefix

if TYPE_CHECKING:
    from wandb_gql import Client


class Visibility(str, Enum):
    # names are what users see/pass into Python methods
    # values are what's expected by backend API
    organization = "PRIVATE"
    restricted = "RESTRICTED"

    @classmethod
    def _missing_(cls, value: object) -> Any:
        # Allow instantiation from enum names too (e.g. "organization" or "restricted")
        return cls.__members__.get(value)

    @classmethod
    def from_gql(cls, value: str) -> Visibility:
        """Convert a GraphQL `visibility` value to a Visibility enum."""
        try:
            return cls(value)
        except ValueError:
            expected = ",".join(repr(e.value) for e in cls)
            raise ValueError(
                f"Invalid visibility {value!r} from backend. Expected one of: {expected}"
            ) from None

    @classmethod
    def from_python(cls, name: str) -> Visibility:
        """Convert a visibility string to a `Visibility` enum."""
        try:
            return cls(name)
        except ValueError:
            expected = ",".join(repr(e.name) for e in cls)
            raise ValueError(
                f"Invalid visibility {name!r}. Expected one of: {expected}"
            ) from None


def prepare_artifact_types_input(
    artifact_types: Collection[str] | None,
) -> list[dict[str, str]] | None:
    """Format the artifact types for the GQL input.

    Args:
        artifact_types: The artifact types to add to the registry.

    Returns:
        The artifact types for the GQL input.
    """
    from wandb.sdk.artifacts._validators import validate_artifact_types

    if artifact_types:
        return [{"name": typ} for typ in validate_artifact_types(artifact_types)]
    return None


def ensure_registry_prefix_on_names(query: Any, in_name: bool = False) -> Any:
    """Recursively the registry prefix to values under "name" keys, excluding regex ops.

    - in_name: True if we are under a "name" key (or propagating from one).

    EX: {"name": "model"} -> {"name": "wandb-registry-model"}
    """
    from wandb.sdk.artifacts._validators import REGISTRY_PREFIX

    if isinstance((txt := query), str):
        return ensureprefix(txt, REGISTRY_PREFIX) if in_name else txt
    if isinstance((dct := query), dict):
        new_dict = {}
        for key, obj in dct.items():
            if key == "$regex":
                # For regex operator, we skip transformation of its value.
                new_dict[key] = obj
            elif key == "name":
                new_dict[key] = ensure_registry_prefix_on_names(obj, in_name=True)
            else:
                # For any other key, propagate flags as-is.
                new_dict[key] = ensure_registry_prefix_on_names(obj, in_name=in_name)
        return new_dict
    if isinstance((seq := query), (list, tuple)):
        return list(map(partial(ensure_registry_prefix_on_names, in_name=in_name), seq))
    return query


@lru_cache(maxsize=10)
def fetch_org_entity_from_organization(client: Client, organization: str) -> str:
    """Fetch the org entity from the organization.

    Args:
        client (Client): Graphql client.
        organization (str): The organization to fetch the org entity for.
    """
    query = gql(
        """
        query FetchOrgEntityFromOrganization($organization: String!) {
            organization(name: $organization) {
                    orgEntity {
                        name
                    }
                }
            }
        """
    )
    try:
        response = client.execute(query, variable_values={"organization": organization})
    except Exception as e:
        raise ValueError(
            f"Error fetching org entity for organization: {organization!r}"
        ) from e

    if (
        not (org := response["organization"])
        or not (org_entity := org["orgEntity"])
        or not (org_name := org_entity["name"])
    ):
        raise ValueError(f"Organization entity for {organization!r} not found.")

    return org_name



"""Module for intercepting stdout/stderr.

This patches the `write()` method of `stdout` and `stderr` on import.
Once patched, it is not possible to unpatch or repatch, though individual
callbacks can be removed.

We assume that all other writing methods on the object delegate to `write()`,
like `writelines()`. This is not guaranteed to be true, but it is true for
common implementations. In particular, CPython's implementation of IOBase's
`writelines()` delegates to `write()`.

It is important to note that this technique interacts poorly with other
code that performs similar patching if it also allows unpatching as this
discards our modification. This is why we patch on import and do not support
unpatching:

    with contextlib.redirect_stderr(...):
        from ... import console_capture
        # Here, everything works fine.
    # Here, callbacks are never called again.

In particular, it does not work with some combinations of pytest's
`capfd` / `capsys` fixtures and pytest's `--capture` option.
"""

from __future__ import annotations

import contextlib
import contextvars
import logging
import sys
import threading
from typing import IO, AnyStr, Callable, Iterator, Protocol

from . import wb_logging

_logger = logging.getLogger(__name__)


class CannotCaptureConsoleError(Exception):
    """The module failed to patch stdout or stderr."""


class _WriteCallback(Protocol):
    """A callback that receives intercepted bytes or string data.

    This may be called from any thread, but is only called from one thread
    at a time.

    Note on errors: Any error raised during the callback will clear all
    callbacks. This means that if a user presses Ctrl-C at an unlucky time
    during a run, we will stop uploading console output---but it's not
    likely to be a problem unless something catches the KeyboardInterrupt.

    Regular Exceptions are caught and logged instead of bubbling up to the
    user's print() statements; other exceptions like KeyboardInterrupt are
    re-raised.

    Callbacks should handle all exceptions---a callback that raises any
    Exception is considered buggy.
    """

    def __call__(
        self,
        data: bytes | str,
        written: int,
        /,
    ) -> None:
        """Intercept data passed to `write()`.

        See the protocol docstring for information about exceptions.

        Args:
            data: The object passed to stderr's or stdout's `write()`.
            written: The number of bytes or characters written.
                This is the return value of `write()`.
        """


_module_lock = threading.Lock()

# See _enter_callbacks().
_is_writing = False
_is_caused_by_callback = contextvars.ContextVar(
    "_is_caused_by_callback",
    default=False,
)

_patch_exception: CannotCaptureConsoleError | None = None

_next_callback_id: int = 1

_stdout_callbacks: dict[int, _WriteCallback] = {}
_stderr_callbacks: dict[int, _WriteCallback] = {}


def capture_stdout(callback: _WriteCallback) -> Callable[[], None]:
    """Install a callback that runs after every write to sys.stdout.

    Args:
        callback: A callback to invoke after running `sys.stdout.write`.

    Returns:
        A function to uninstall the callback.

    Raises:
        CannotCaptureConsoleError: If patching failed on import.
    """
    with _module_lock:
        if _patch_exception:
            raise _patch_exception

        return _insert_disposably(
            _stdout_callbacks,
            callback,
        )


def capture_stderr(callback: _WriteCallback) -> Callable[[], None]:
    """Install a callback that runs after every write to sys.sdterr.

    Args:
        callback: A callback to invoke after running `sys.stderr.write`.

    Returns:
        A function to uninstall the callback.

    Raises:
        CannotCaptureConsoleError: If patching failed on import.
    """
    with _module_lock:
        if _patch_exception:
            raise _patch_exception

        return _insert_disposably(
            _stderr_callbacks,
            callback,
        )


def _insert_disposably(
    callback_dict: dict[int, _WriteCallback],
    callback: _WriteCallback,
) -> Callable[[], None]:
    global _next_callback_id
    id = _next_callback_id
    _next_callback_id += 1

    disposed = False

    def dispose() -> None:
        nonlocal disposed

        with _module_lock:
            if disposed:
                return

            callback_dict.pop(id, None)

            disposed = True

    callback_dict[id] = callback
    return dispose


def _patch(
    stdout_or_stderr: IO[AnyStr],
    callbacks: dict[int, _WriteCallback],
) -> None:
    orig_write: Callable[[AnyStr], int]

    def write_with_callbacks(s: AnyStr, /) -> int:
        n = orig_write(s)

        with contextlib.ExitStack() as stack:
            stack.enter_context(_reset_on_exception())
            stack.enter_context(wb_logging.log_to_all_runs())
            callbacks_list = stack.enter_context(_enter_callbacks(callbacks))

            for cb in callbacks_list:
                cb(s, n)

        return n

    orig_write = stdout_or_stderr.write

    # mypy==1.14.1 fails to type-check this:
    #   Incompatible types in assignment (expression has type
    #   "Callable[[bytes], int]", variable has type overloaded function)
    stdout_or_stderr.write = write_with_callbacks  # type: ignore


@contextlib.contextmanager
def _enter_callbacks(
    callbacks: dict[int, _WriteCallback],
) -> Iterator[list[_WriteCallback]]:
    """Returns a list of callbacks to invoke.

    This prevents deadlocks and some infinite loops by returning an empty list
    when:

    * A callback prints
    * A callback blocks on a thread that's printing
    * A callback schedules an async task that prints

    It is impossible to prevent all infinite loops: a callback could put
    a message into a queue, causing an unrelated thread to print later,
    invoking the same callback and repeating forever.
    """
    global _is_writing

    # The global _is_writing variable is necessary despite the contextvar
    # because it's possible to create a thread without copying the context.
    # This is the default behavior for threading.Thread() before Python 3.14.
    #
    # A side effect of it is that when multiple threads print simultaneously,
    # some messages will not be captured.

    with _module_lock:
        if _is_writing or _is_caused_by_callback.get():
            callbacks_list = None

        else:
            callbacks_list = list(callbacks.values())
            _is_writing = True
            _is_caused_by_callback.set(True)

    if callbacks_list is None:
        yield []
        return

    try:
        yield callbacks_list
    finally:
        with _module_lock:
            _is_writing = False
            _is_caused_by_callback.set(False)


@contextlib.contextmanager
def _reset_on_exception() -> Iterator[None]:
    """Clear all callbacks on any exception, suppressing it.

    This prevents infinite loops:

    * If we re-raise, an exception handler is likely to print
      the exception to the console and trigger callbacks again
    * If we log, we can't guarantee that this doesn't print
      to console.

    This is especially important for KeyboardInterrupt.
    """
    try:
        yield
    except BaseException as e:
        with _module_lock:
            _stderr_callbacks.clear()
            _stdout_callbacks.clear()

        if isinstance(e, Exception):
            # We suppress Exceptions so that bugs in W&B code don't
            # cause the user's print() statements to raise errors.
            _logger.exception("Error in console callback, clearing all!")
        else:
            # Re-raise errors like KeyboardInterrupt.
            raise


try:
    _patch(sys.stdout, _stdout_callbacks)
    _patch(sys.stderr, _stderr_callbacks)
except Exception as _patch_exception_cause:
    _patch_exception = CannotCaptureConsoleError()
    _patch_exception.__cause__ = _patch_exception_cause



"""Validation for URLs."""

import re

from wandb._pydantic import IS_PYDANTIC_V2


def validate_url(url: object) -> None:
    """Validate a URL.

    Args:
        url: The URL to validate.

    Raises:
        ValueError: If the URL is invalid.
        TypeError: If given something other than a string.
    """
    if not isinstance(url, str):
        raise TypeError(f"Expected a string, got {type(url)}")

    if IS_PYDANTIC_V2:
        _validate_url_pydantic(url)
    else:
        _validate_url_custom(url)


def _validate_url_pydantic(url: str) -> None:
    """Validate a URL using Pydantic's validator."""
    from pydantic_core import SchemaValidator, core_schema

    SchemaValidator(
        core_schema.url_schema(
            allowed_schemes=["http", "https"],
            strict=True,
        )
    ).validate_python(url)


def _validate_url_custom(url: str) -> None:
    """Validate a URL.

    We will remove this once we can require Pydantic V2.

    Based on the Django URLValidator, but with a few additional checks.

    Copyright (c) Django Software Foundation and individual contributors.
    All rights reserved.

    Redistribution and use in source and binary forms, with or without modification,
    are permitted provided that the following conditions are met:

        1. Redistributions of source code must retain the above copyright notice,
            this list of conditions and the following disclaimer.

        2. Redistributions in binary form must reproduce the above copyright
            notice, this list of conditions and the following disclaimer in the
            documentation and/or other materials provided with the distribution.

        3. Neither the name of Django nor the names of its contributors may be used
            to endorse or promote products derived from this software without
            specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
    ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
    WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
    ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
    (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
    ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
    """
    from urllib.parse import urlparse, urlsplit

    ul = "\u00a1-\uffff"  # Unicode letters range (must not be a raw string).

    # IP patterns
    ipv4_re = (
        r"(?:0|25[0-5]|2[0-4][0-9]|1[0-9]?[0-9]?|[1-9][0-9]?)"
        r"(?:\.(?:0|25[0-5]|2[0-4][0-9]|1[0-9]?[0-9]?|[1-9][0-9]?)){3}"
    )
    ipv6_re = r"\[[0-9a-f:.]+\]"  # (simple regex, validated later)

    # Host patterns
    hostname_re = (
        r"[a-z" + ul + r"0-9](?:[a-z" + ul + r"0-9-]{0,61}[a-z" + ul + r"0-9])?"
    )
    # Max length for domain name labels is 63 characters per RFC 1034 sec. 3.1
    domain_re = r"(?:\.(?!-)[a-z" + ul + r"0-9-]{1,63}(?<!-))*"
    tld_re = (
        r"\."  # dot
        r"(?!-)"  # can't start with a dash
        r"(?:[a-z" + ul + "-]{2,63}"  # domain label
        r"|xn--[a-z0-9]{1,59})"  # or punycode label
        r"(?<!-)"  # can't end with a dash
        r"\.?"  # may have a trailing dot
    )
    # host_re = "(" + hostname_re + domain_re + tld_re + "|localhost)"
    # todo?: allow hostname to be just a hostname (no tld)?
    host_re = "(" + hostname_re + domain_re + f"({tld_re})?" + "|localhost)"

    regex = re.compile(
        r"^(?:[a-z0-9.+-]*)://"  # scheme is validated separately
        r"(?:[^\s:@/]+(?::[^\s:@/]*)?@)?"  # user:pass authentication
        r"(?:" + ipv4_re + "|" + ipv6_re + "|" + host_re + ")"
        r"(?::[0-9]{1,5})?"  # port
        r"(?:[/?#][^\s]*)?"  # resource path
        r"\Z",
        re.IGNORECASE,
    )
    schemes = {"http", "https"}
    unsafe_chars = frozenset("\t\r\n")

    scheme = url.split("://")[0].lower()
    split_url = urlsplit(url)
    parsed_url = urlparse(url)

    if parsed_url.netloc == "":
        raise ValueError(f"Invalid URL: {url!r}")
    elif unsafe_chars.intersection(url):
        raise ValueError("URL cannot contain unsafe characters")
    elif scheme not in schemes:
        raise ValueError("URL must start with `http(s)://`")
    elif not regex.search(url):
        raise ValueError(f"{url!r} is not a valid server address")
    elif split_url.hostname is None or len(split_url.hostname) > 253:
        raise ValueError("hostname is invalid")



#
"""InternalRun - Internal-only run object.

Semi-stubbed run for internal process use.

"""

from typing_extensions import override

from wandb.sdk import wandb_run


class InternalRun(wandb_run.Run):
    def __init__(self, run_obj, settings, datatypes_cb):
        super().__init__(settings=settings)
        self._run_obj = run_obj
        self._datatypes_cb = datatypes_cb

    @override
    def _set_backend(self, backend):
        # This type of run object can't have a backend
        # or do any writes.
        pass

    @override
    def _publish_file(self, fname: str) -> None:
        self._datatypes_cb(fname)



# Generated by ariadne-codegen

__all__ = [
    "ADD_ALIASES_GQL",
    "ADD_ARTIFACT_COLLECTION_TAGS_GQL",
    "ARTIFACT_BY_ID_GQL",
    "ARTIFACT_BY_NAME_GQL",
    "ARTIFACT_COLLECTION_ALIASES_GQL",
    "ARTIFACT_CREATED_BY_GQL",
    "ARTIFACT_MEMBERSHIP_BY_NAME_GQL",
    "ARTIFACT_TYPE_GQL",
    "ARTIFACT_USED_BY_GQL",
    "CREATE_REGISTRY_MEMBERS_GQL",
    "DELETE_ALIASES_GQL",
    "DELETE_ARTIFACT_COLLECTION_TAGS_GQL",
    "DELETE_ARTIFACT_GQL",
    "DELETE_ARTIFACT_PORTFOLIO_GQL",
    "DELETE_ARTIFACT_SEQUENCE_GQL",
    "DELETE_REGISTRY_GQL",
    "DELETE_REGISTRY_MEMBERS_GQL",
    "FETCH_ARTIFACT_MANIFEST_GQL",
    "FETCH_LINKED_ARTIFACTS_GQL",
    "FETCH_ORG_INFO_FROM_ENTITY_GQL",
    "FETCH_REGISTRIES_GQL",
    "FETCH_REGISTRY_GQL",
    "GET_ARTIFACT_FILES_GQL",
    "GET_ARTIFACT_FILE_URLS_GQL",
    "GET_ARTIFACT_MEMBERSHIP_FILES_GQL",
    "GET_ARTIFACT_MEMBERSHIP_FILE_URLS_GQL",
    "LINK_ARTIFACT_GQL",
    "PROJECT_ARTIFACTS_GQL",
    "PROJECT_ARTIFACT_COLLECTIONS_GQL",
    "PROJECT_ARTIFACT_COLLECTION_GQL",
    "PROJECT_ARTIFACT_TYPES_GQL",
    "PROJECT_ARTIFACT_TYPE_GQL",
    "REGISTRY_COLLECTIONS_GQL",
    "REGISTRY_TEAM_MEMBERS_GQL",
    "REGISTRY_USER_MEMBERS_GQL",
    "REGISTRY_VERSIONS_GQL",
    "RENAME_REGISTRY_GQL",
    "RUN_INPUT_ARTIFACTS_GQL",
    "RUN_OUTPUT_ARTIFACTS_GQL",
    "TYPE_INFO_GQL",
    "UNLINK_ARTIFACT_GQL",
    "UPDATE_ARTIFACT_GQL",
    "UPDATE_ARTIFACT_PORTFOLIO_GQL",
    "UPDATE_ARTIFACT_SEQUENCE_GQL",
    "UPDATE_ARTIFACT_SEQUENCE_TYPE_GQL",
    "UPDATE_TEAM_REGISTRY_ROLE_GQL",
    "UPDATE_USER_REGISTRY_ROLE_GQL",
    "UPSERT_REGISTRY_GQL",
    "DeleteArtifactSequence",
    "DeleteArtifactPortfolio",
    "UpdateArtifactSequence",
    "UpdateArtifactPortfolio",
    "UpdateArtifactSequenceType",
    "AddArtifactCollectionTags",
    "DeleteArtifactCollectionTags",
    "ProjectArtifactCollections",
    "ProjectArtifactCollection",
    "ArtifactCollectionAliases",
    "GetArtifactFiles",
    "GetArtifactMembershipFiles",
    "GetArtifactFileUrls",
    "GetArtifactMembershipFileUrls",
    "ProjectArtifactTypes",
    "ProjectArtifactType",
    "ProjectArtifacts",
    "RunOutputArtifacts",
    "RunInputArtifacts",
    "FetchLinkedArtifacts",
    "FetchArtifactManifest",
    "ArtifactByID",
    "ArtifactByName",
    "ArtifactMembershipByName",
    "ArtifactUsedBy",
    "ArtifactCreatedBy",
    "ArtifactType",
    "AddAliases",
    "DeleteAliases",
    "UpdateArtifact",
    "DeleteArtifact",
    "LinkArtifact",
    "UnlinkArtifact",
    "TypeInfo",
    "FetchOrgInfoFromEntity",
    "RegistryVersions",
    "RegistryCollections",
    "FetchRegistry",
    "FetchRegistries",
    "RenameRegistry",
    "UpsertRegistry",
    "DeleteRegistry",
    "RegistryUserMembers",
    "RegistryTeamMembers",
    "CreateRegistryMembers",
    "DeleteRegistryMembers",
    "UpdateUserRegistryRole",
    "UpdateTeamRegistryRole",
    "AddAliasesInput",
    "ArtifactAliasInput",
    "ArtifactCollectionAliasInput",
    "ArtifactTypeInput",
    "CreateArtifactCollectionTagAssignmentsInput",
    "CreateProjectMembersInput",
    "DeleteAliasesInput",
    "DeleteArtifactCollectionTagAssignmentsInput",
    "DeleteArtifactInput",
    "DeleteProjectMembersInput",
    "LinkArtifactInput",
    "MoveArtifactSequenceInput",
    "RateLimitsInput",
    "RenameProjectInput",
    "TagInput",
    "UnlinkArtifactInput",
    "UpdateArtifactInput",
    "UpdateArtifactPortfolioInput",
    "UpdateArtifactSequenceInput",
    "UpdateProjectMemberInput",
    "UpdateProjectTeamMemberInput",
    "UpsertModelInput",
    "ArtifactAliasFragment",
    "ArtifactCollectionFragment",
    "ArtifactFragment",
    "ArtifactMembershipFragment",
    "ArtifactPortfolioTypeFields",
    "ArtifactSequenceTypeFields",
    "ArtifactTypeFragment",
    "CollectionInfoFragment",
    "DeferredManifestFragment",
    "FileFragment",
    "FileWithUrlFragment",
    "OrgInfoFragment",
    "PageInfoFragment",
    "ProjectInfoFragment",
    "RegistryCollectionFragment",
    "RegistryFragment",
    "RegistryRoleFragment",
    "RunInfoFragment",
    "SourceCollectionInfoFragment",
    "TagFragment",
    "TeamFragment",
    "TeamMemberFragment",
    "TeamRegistryMemberFragment",
    "TypeInfoFragment",
    "UserRegistryMemberFragment",
    "ArtifactCollectionState",
    "ArtifactCollectionType",
    "ArtifactState",
]
from .add_aliases import AddAliases
from .add_artifact_collection_tags import AddArtifactCollectionTags
from .artifact_by_id import ArtifactByID
from .artifact_by_name import ArtifactByName
from .artifact_collection_aliases import ArtifactCollectionAliases
from .artifact_created_by import ArtifactCreatedBy
from .artifact_membership_by_name import ArtifactMembershipByName
from .artifact_type import ArtifactType
from .artifact_used_by import ArtifactUsedBy
from .create_registry_members import CreateRegistryMembers
from .delete_aliases import DeleteAliases
from .delete_artifact import DeleteArtifact
from .delete_artifact_collection_tags import DeleteArtifactCollectionTags
from .delete_artifact_portfolio import DeleteArtifactPortfolio
from .delete_artifact_sequence import DeleteArtifactSequence
from .delete_registry import DeleteRegistry
from .delete_registry_members import DeleteRegistryMembers
from .enums import ArtifactCollectionState, ArtifactCollectionType, ArtifactState
from .fetch_artifact_manifest import FetchArtifactManifest
from .fetch_linked_artifacts import FetchLinkedArtifacts
from .fetch_org_info_from_entity import FetchOrgInfoFromEntity
from .fetch_registries import FetchRegistries
from .fetch_registry import FetchRegistry
from .fragments import (
    ArtifactAliasFragment,
    ArtifactCollectionFragment,
    ArtifactFragment,
    ArtifactMembershipFragment,
    ArtifactPortfolioTypeFields,
    ArtifactSequenceTypeFields,
    ArtifactTypeFragment,
    CollectionInfoFragment,
    DeferredManifestFragment,
    FileFragment,
    FileWithUrlFragment,
    OrgInfoFragment,
    PageInfoFragment,
    ProjectInfoFragment,
    RegistryCollectionFragment,
    RegistryFragment,
    RegistryRoleFragment,
    RunInfoFragment,
    SourceCollectionInfoFragment,
    TagFragment,
    TeamFragment,
    TeamMemberFragment,
    TeamRegistryMemberFragment,
    TypeInfoFragment,
    UserRegistryMemberFragment,
)
from .get_artifact_file_urls import GetArtifactFileUrls
from .get_artifact_files import GetArtifactFiles
from .get_artifact_membership_file_urls import GetArtifactMembershipFileUrls
from .get_artifact_membership_files import GetArtifactMembershipFiles
from .input_types import (
    AddAliasesInput,
    ArtifactAliasInput,
    ArtifactCollectionAliasInput,
    ArtifactTypeInput,
    CreateArtifactCollectionTagAssignmentsInput,
    CreateProjectMembersInput,
    DeleteAliasesInput,
    DeleteArtifactCollectionTagAssignmentsInput,
    DeleteArtifactInput,
    DeleteProjectMembersInput,
    LinkArtifactInput,
    MoveArtifactSequenceInput,
    RateLimitsInput,
    RenameProjectInput,
    TagInput,
    UnlinkArtifactInput,
    UpdateArtifactInput,
    UpdateArtifactPortfolioInput,
    UpdateArtifactSequenceInput,
    UpdateProjectMemberInput,
    UpdateProjectTeamMemberInput,
    UpsertModelInput,
)
from .link_artifact import LinkArtifact
from .operations import (
    ADD_ALIASES_GQL,
    ADD_ARTIFACT_COLLECTION_TAGS_GQL,
    ARTIFACT_BY_ID_GQL,
    ARTIFACT_BY_NAME_GQL,
    ARTIFACT_COLLECTION_ALIASES_GQL,
    ARTIFACT_CREATED_BY_GQL,
    ARTIFACT_MEMBERSHIP_BY_NAME_GQL,
    ARTIFACT_TYPE_GQL,
    ARTIFACT_USED_BY_GQL,
    CREATE_REGISTRY_MEMBERS_GQL,
    DELETE_ALIASES_GQL,
    DELETE_ARTIFACT_COLLECTION_TAGS_GQL,
    DELETE_ARTIFACT_GQL,
    DELETE_ARTIFACT_PORTFOLIO_GQL,
    DELETE_ARTIFACT_SEQUENCE_GQL,
    DELETE_REGISTRY_GQL,
    DELETE_REGISTRY_MEMBERS_GQL,
    FETCH_ARTIFACT_MANIFEST_GQL,
    FETCH_LINKED_ARTIFACTS_GQL,
    FETCH_ORG_INFO_FROM_ENTITY_GQL,
    FETCH_REGISTRIES_GQL,
    FETCH_REGISTRY_GQL,
    GET_ARTIFACT_FILE_URLS_GQL,
    GET_ARTIFACT_FILES_GQL,
    GET_ARTIFACT_MEMBERSHIP_FILE_URLS_GQL,
    GET_ARTIFACT_MEMBERSHIP_FILES_GQL,
    LINK_ARTIFACT_GQL,
    PROJECT_ARTIFACT_COLLECTION_GQL,
    PROJECT_ARTIFACT_COLLECTIONS_GQL,
    PROJECT_ARTIFACT_TYPE_GQL,
    PROJECT_ARTIFACT_TYPES_GQL,
    PROJECT_ARTIFACTS_GQL,
    REGISTRY_COLLECTIONS_GQL,
    REGISTRY_TEAM_MEMBERS_GQL,
    REGISTRY_USER_MEMBERS_GQL,
    REGISTRY_VERSIONS_GQL,
    RENAME_REGISTRY_GQL,
    RUN_INPUT_ARTIFACTS_GQL,
    RUN_OUTPUT_ARTIFACTS_GQL,
    TYPE_INFO_GQL,
    UNLINK_ARTIFACT_GQL,
    UPDATE_ARTIFACT_GQL,
    UPDATE_ARTIFACT_PORTFOLIO_GQL,
    UPDATE_ARTIFACT_SEQUENCE_GQL,
    UPDATE_ARTIFACT_SEQUENCE_TYPE_GQL,
    UPDATE_TEAM_REGISTRY_ROLE_GQL,
    UPDATE_USER_REGISTRY_ROLE_GQL,
    UPSERT_REGISTRY_GQL,
)
from .project_artifact_collection import ProjectArtifactCollection
from .project_artifact_collections import ProjectArtifactCollections
from .project_artifact_type import ProjectArtifactType
from .project_artifact_types import ProjectArtifactTypes
from .project_artifacts import ProjectArtifacts
from .registry_collections import RegistryCollections
from .registry_team_members import RegistryTeamMembers
from .registry_user_members import RegistryUserMembers
from .registry_versions import RegistryVersions
from .rename_registry import RenameRegistry
from .run_input_artifacts import RunInputArtifacts
from .run_output_artifacts import RunOutputArtifacts
from .type_info import TypeInfo
from .unlink_artifact import UnlinkArtifact
from .update_artifact import UpdateArtifact
from .update_artifact_portfolio import UpdateArtifactPortfolio
from .update_artifact_sequence import UpdateArtifactSequence
from .update_artifact_sequence_type import UpdateArtifactSequenceType
from .update_team_registry_role import UpdateTeamRegistryRole
from .update_user_registry_role import UpdateUserRegistryRole
from .upsert_registry import UpsertRegistry



from __future__ import annotations

import os
import secrets
import socket
import string

import wandb

from . import config
from . import files as sm_files


def set_run_id(run_settings: wandb.Settings) -> bool:
    """Set a run ID and group when using SageMaker.

    Returns whether the ID and group were updated.
    """
    # Added in https://github.com/wandb/wandb/pull/3290.
    #
    # Prevents SageMaker from overriding the run ID configured
    # in environment variables. Note, however, that it will still
    # override a run ID passed explicitly to `wandb.init()`.
    if os.getenv("WANDB_RUN_ID"):
        return False

    run_group = os.getenv("TRAINING_JOB_NAME")
    if not run_group:
        return False

    alphanumeric = string.ascii_lowercase + string.digits
    random = "".join(secrets.choice(alphanumeric) for _ in range(6))

    host = os.getenv("CURRENT_HOST", socket.gethostname())

    run_settings.run_id = f"{run_group}-{random}-{host}"
    run_settings.run_group = run_group
    return True


def set_global_settings(settings: wandb.Settings) -> None:
    """Set global W&B settings based on the SageMaker environment."""
    if env := parse_sm_secrets():
        settings.update_from_env_vars(env)

    # The SageMaker config may contain an API key, in which case it
    # takes precedence over the value in the secrets. It's unclear
    # whether this is by design, or by accident; we keep it for
    # backward compatibility for now.
    sm_config = config.parse_sm_config()
    if api_key := sm_config.get("wandb_api_key"):
        settings.api_key = api_key


def parse_sm_secrets() -> dict[str, str]:
    """We read our api_key from secrets.env in SageMaker."""
    env_dict = dict()
    # Set secret variables
    if os.path.exists(sm_files.SM_SECRETS):
        for line in open(sm_files.SM_SECRETS):
            key, val = line.strip().split("=", 1)
            env_dict[key] = val
    return env_dict



"""W&B Integration for Metaflow.

Defines a custom step and flow decorator `wandb_log` that automatically logs
flow parameters and artifacts to W&B.
"""

from .metaflow import wandb_log, wandb_track, wandb_use

__all__ = ["wandb_log", "wandb_track", "wandb_use"]



# Generated by ariadne-codegen
# Source: tools/graphql_codegen/artifacts/

from __future__ import annotations

from typing import Optional

from pydantic import Field

from wandb._pydantic import GQLResult

from .fragments import ArtifactCollectionFragment


class ProjectArtifactCollection(GQLResult):
    project: Optional[ProjectArtifactCollectionProject]


class ProjectArtifactCollectionProject(GQLResult):
    artifact_type: Optional[ProjectArtifactCollectionProjectArtifactType] = Field(
        alias="artifactType"
    )


class ProjectArtifactCollectionProjectArtifactType(GQLResult):
    artifact_collection: Optional[ArtifactCollectionFragment] = Field(
        alias="artifactCollection"
    )


ProjectArtifactCollection.model_rebuild()
ProjectArtifactCollectionProject.model_rebuild()
ProjectArtifactCollectionProjectArtifactType.model_rebuild()



# Generated by ariadne-codegen
# Source: tools/graphql_codegen/artifacts/

from __future__ import annotations

from typing import List, Optional

from pydantic import Field

from wandb._pydantic import GQLResult

from .fragments import ArtifactCollectionFragment, PageInfoFragment


class ProjectArtifactCollections(GQLResult):
    project: Optional[ProjectArtifactCollectionsProject]


class ProjectArtifactCollectionsProject(GQLResult):
    artifact_type: Optional[ProjectArtifactCollectionsProjectArtifactType] = Field(
        alias="artifactType"
    )


class ProjectArtifactCollectionsProjectArtifactType(GQLResult):
    artifact_collections: Optional[
        ProjectArtifactCollectionsProjectArtifactTypeArtifactCollections
    ] = Field(alias="artifactCollections")


class ProjectArtifactCollectionsProjectArtifactTypeArtifactCollections(GQLResult):
    total_count: int = Field(alias="totalCount")
    page_info: PageInfoFragment = Field(alias="pageInfo")
    edges: List[ProjectArtifactCollectionsProjectArtifactTypeArtifactCollectionsEdges]


class ProjectArtifactCollectionsProjectArtifactTypeArtifactCollectionsEdges(GQLResult):
    node: Optional[ArtifactCollectionFragment]


ProjectArtifactCollections.model_rebuild()
ProjectArtifactCollectionsProject.model_rebuild()
ProjectArtifactCollectionsProjectArtifactType.model_rebuild()
ProjectArtifactCollectionsProjectArtifactTypeArtifactCollections.model_rebuild()
ProjectArtifactCollectionsProjectArtifactTypeArtifactCollectionsEdges.model_rebuild()



"""Pydantic model classes and related helpers for artifacts code.

Excludes GraphQL-generated classes.
"""

__all__ = [
    "ArtifactsBase",
    "RegistryData",
    "ArtifactCollectionData",
]
from .artifact_collection import ArtifactCollectionData
from .base_model import ArtifactsBase
from .registry import RegistryData



# Generated by ariadne-codegen
# Source: tools/graphql_codegen/artifacts/

from __future__ import annotations

from typing import List, Optional

from pydantic import Field

from wandb._pydantic import GQLResult

from .fragments import RunInfoFragment


class ArtifactUsedBy(GQLResult):
    artifact: Optional[ArtifactUsedByArtifact]


class ArtifactUsedByArtifact(GQLResult):
    used_by: ArtifactUsedByArtifactUsedBy = Field(alias="usedBy")


class ArtifactUsedByArtifactUsedBy(GQLResult):
    edges: List[ArtifactUsedByArtifactUsedByEdges]


class ArtifactUsedByArtifactUsedByEdges(GQLResult):
    node: RunInfoFragment


ArtifactUsedBy.model_rebuild()
ArtifactUsedByArtifact.model_rebuild()
ArtifactUsedByArtifactUsedBy.model_rebuild()
ArtifactUsedByArtifactUsedByEdges.model_rebuild()
